/* Copyright (c) 2000-2003 ActiveState Corp.
   See the file LICENSE.txt for licensing information. */

/* The new new Rx dialog.
 *
 * See:
 *  http://www.perl.com/doc/manual/html/pod/perlre.html
 *  http://www.perldoc.com/perl5.6/pod/perlre.html
 *  XXX Should link to ASPN version?
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0].
 *      .language   The regular expression language syntax to use. Defaults to
 *                  Python (XXX for now, but will likely default to Perl).
 *      .regex      A regular expression to use. Defaults to the selection in
 *                  the current buffer (if there is one) or the regex value
 *                  when Rx was last closed.
 *      .replacement  A replacement pattern to use.
 *      .searchText
 *      .mode
 *
 *  On return window.arguments[0] has:
 *      .language   The regular expression language syntax set at exit.
 *      .regex      The regular expression set at exit.
 *      .replacement  The replacement pattern at exit.
 *      .searchText
 *      .mode
 */
//---- Dev Notes:
// What modes to have?
//  - Python: match, search, findall, replace, split
//  - Perl: match (like Python's search), replace, split, ['g' modifier]
//  - Perl alternative: match, match all, replace, replace all, split
//  - PHP: (possibly grep), match, match all, replace, split
//  - Tcl: match, replace, (possibly match all, replace all)
//  - Blue Sky for all: replace with callback
// Solution:
//    <match> <match all> <split> <replace> <replace all>
// Python (for now at least) will not have a mode for re.match(). Note that
// the "match" mode is akin to Python's re.search().
//
//---- BUGS:
//  - <Tab> does nothing in any of the scintillas
//  - typing in Search Text does not update highlighting of hits
//  - When jumping to a match result (by double clicking on the "folder" icon
//    for it), the event is not cancelled and a result container will be
//    opened or closed. (Need help on this one.)
//
//---- PERF:
//  - PERF: need to do match results updating and search text styling only
//    every second or so, e.g. regex="foo|" which matches at _every_ index.
//
//---- TODO:
//  - drag and drop support: try to drag text around to get tracebacks
//  - icons for the match results tree view
//  - put little grippies on the splitters
//  - NOTE additional PHP regex modifiers:  http://ca3.php.net/manual/en/pcre.pattern.modifiers.php
//  - perlre.html describes a case where regex matches can yield Perl warnings.
//    We should trap and show those.
//  - The perl backend has to know whether to "use utf8;" and "use
//    <someotherone>;". Read perlre.html for details.
//  - add all the Perl appropriate shortcuts
//  - implement extracting regex from the current
//    document (will this require heuristics for
//    Python? how to tell if a Python string is a
//    string or a regex? should Rxx take the hint
//    it cursor is on a re.compile or re.search,
//    etc. method?)
//  - implement ties to a current view, i.e. (1)
//    inserting the result back in, (2) others?
//  - implement for Perl
//  - implement providing a documented/verbosified version of
//    the regex
//  - mru menu (or should this be an explicit save list?)
//    Could just ask (w/ don't ask again) with a dialog on
//    exit, each time.
//  - "describe" button to pop a dialog with a verbose layout
//    of the regex, this dialog offers to replace the existing
//    one?
//  - general scintilla buffer properties?
//  - help menu could include link to Rx Cookbook on ASPN
//  - try using the INDIC_BOX indicator to mark matches instead of the
//    of background color highlighting when we upgrade to the Scintilla
//    that supports this
//  - Fix double-click (select word) in regex view to recognize lexical state
//    boundaries such that double-clicking on "foo" in the following, just
//    selected "foo":
//        \x0Ffoo
//  - Would be nice to have Ctrl+<right/left-arrow> also use lexical states
//    to effect.
//
// Blue Sky Feature Requests:
//  - Animate the hiding/showing of the replace column
//    (see _Rxx_{Show|Hide}ReplaceColumn).
//  - Vim, egrep, grep, Emacs regex syntaxes
//  - Highlighting feature: shade the background of groups and make the
//    shading darker the more deeply nested the group.
//  - Consider an outliner view of the regex structure (with a tree with
//    hierarchy), similar to what OptiPerl has.
//  - Perl tr///cds support


var log = getLoggingMgr().getLogger("rxx");
//log.setLevel(LOG_DEBUG);

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const SC_MOD_INSERTTEXT = Components.interfaces.ISciMoz.SC_MOD_INSERTTEXT;
const SC_MOD_DELETETEXT = Components.interfaces.ISciMoz.SC_MOD_DELETETEXT;

var gPrefs = null;  // access to global prefs
var gWidgets = null; // cache of widget DOM nodes
var gRegexEventist = null; // event handler for the regex SciMoz
var gReplacementEventist = null; // event handler for the replacement SciMoz
var gSearchTextEventist = null; // event handler for the search text SciMoz
var gReplacedTextEventist = null; // event handler for the replaced text SciMoz
// Boolean indicating if the regex syntax shortcuts menupopup needs to be
// (re)built.
var gNeedToRebuildRegexShortcuts = null;
var gNeedToRebuildReplacementShortcuts = null;
var gLanguage = null;
var gMode = null; // e.g. "match", "match-all", "split", "replace", ...
var gRegexOptions = null;
var gHandler = null; // koIRxxHandler instance for the current language
var gResultsManager = null; // JS-implemented XPCOM interface from Python code to JS ui.


var gLanguageToHandlerCID = {
    //"perl": "@activestate.com/koRxxPerlHandler;1",
    "python": "@activestate.com/koRxxPythonHandler;1"
}

var gLanguageToDisplayName = {
    //"perl": "Perl",
    "python": "Python"
}

var gSampleRegex = { // map <language> to sample regex
    //XXX While we only have Python (acting for Perl as well) lets not
    //    use Python-only regex stuff.
    //"python": "(?P<file>.+?):(?P<line>\\d+):(?P<code>.*)",
    "*": "(.+?):(\\d+):(.*)"
}
var gSampleReplacement = { // map <language> to sample replacement pattern
    //XXX While we only have Python (acting for Perl as well) lets not
    //    use Python-only regex stuff.
    //"python": 'file:\\g<file>, line:\\2, code:"\\g<code>"',
    "*": 'file:\\1, line:\\2, code:"\\3"'
}
var gSampleSearchText = {  // map <language> to sample search text
    "*": "foo.pl:25:use strict;\n" +
         "bar.pm:42:sub bar { # womba womba womba\n" +
         "spam.py:234:seuss = ['green', 'eggs', 'ham']\n"
}

var gModeToName = {
    "match": "Match",
    "match-all": "Match All",
    "split": "Split",
    "replace": "Replace",
    "replace-all": "Replace All"
}

var gRegexShortcuts = {
    "python": [
        {"shortcut": ".", "desc": "any character except a newline"},
        {"shortcut": "\\d", "desc": "any decimal digit",
         "tooltip": "equivalent to [0-9]"},
        {"shortcut": "\\D", "desc": "any non-digit",
         "tooltip": "equivalent to [^0-9]"},
        {"shortcut": "\\s", "desc": "any whitespace character",
         "tooltip": "equivalent to [ \\t\\n\\r\\f\\v]"},
        {"shortcut": "\\S", "desc": "any non-whitespace character",
         "tooltip": "equivalent to [^ \\t\\n\\r\\f\\v]"},
        {"shortcut": "\\w", "desc": "any alphanumeric character"},
        {"shortcut": "\\W", "desc": "any non-alphanumeric character"},
        {"shortcut": "\\number",
         "desc": "the contents of the group of the given number"},
        {"shortcut": "(?P=name)", "inserter": "(?P=<<name>>)",
         "desc": "the contents of the group of the given name"},
        {"separator": null},
        {"shortcut": "*", "desc": "zero or more of the preceding block"},
        {"shortcut": "*?", "desc": "zero or more of the preceding block (non-greedy)"},
        {"shortcut": "+", "desc": "one or more of the preceding block"},
        {"shortcut": "+?", "desc": "one or more of the preceding block (non-greedy)"},
        {"shortcut": "?", "desc": "zero or one of the preceding block"},
        {"shortcut": "??", "desc": "zero or one of the preceding block (non-greedy)"},
        {"shortcut": "{m}", "inserter": "{<<m>>}",
         "desc": "exactly 'm' copies of the preceding block"},
        {"shortcut": "{m,n}", "desc": "'m' to 'n' copies of the preceding block"},
        {"shortcut": "{m,n}?", "desc": "'m' to 'n' copies of the preceding block (non-greedy)"},
        {"separator": null},
        {"shortcut": "^", "desc": "beginning of line"},
        {"shortcut": "$", "desc": "end of line"},
        {"shortcut": "\\b", "desc": "the beginning or end of a word"},
        {"shortcut": "\\B", "desc": "anything BUT the beginning or end of a word"},
        {"shortcut": "\\A", "desc": "beginning of the string"},
        {"shortcut": "\\Z", "desc": "end of the string"},
        {"separator": null},
        {"shortcut": "(...)", "desc": "group"},
        {"shortcut": "(?P<name>...)", "desc": "named group"},
        {"shortcut": "(?:...)", "desc": "non-grouping group",
         "tooltip": "the result of the match is NOT assigned a group number"},
        {"shortcut": "(?#...)", "desc": "a comment, the contents are ignored"},
        {"shortcut": "(?=...)",
         "desc": "matches if <...> matches next, but does not consume the string"},
        {"shortcut": "(?!...)",
         "desc": "matches if <...> does NOT match next, but does not consume the string"},
        {"shortcut": "(?<=...)", "desc": "positive lookbehind assertion",
         "tooltip": "matches if the current position is immediately "+
                    "preceded by a match for <...>"},
        {"shortcut": "(?<!...)", "desc": "negative lookbehind assertion",
         "tooltip": "matches if the current position is NOT immediately "+
                    "preceded by a match for <...>"},

        //XXX add entry to "Escape the selection"
    ],
    "perl": [
        {"shortcut": "^", "desc": "beginning of line"},
        {"shortcut": "$", "desc": "end of line"}
    ]
}


//---- the Rxx results manager

function RxxResults_Manager()
{
    this.matchResultsView = Components.classes["@activestate.com/koRxxMatchResultsTreeView;1"]
                              .createInstance();
    gWidgets.matchResultsTree.treeBoxObject.view = this.matchResultsView;
    this.splitResultsView = Components.classes["@activestate.com/koRxxSplitResultsTreeView;1"]
                              .createInstance();
    gWidgets.splitResultsTree.treeBoxObject.view = this.splitResultsView;
}
RxxResults_Manager.prototype.constructor = RxxResults_Manager;

RxxResults_Manager.prototype.QueryInterface = function (iid) {
    if (!iid.equals(Components.interfaces.koIRxxResultsManager) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}

RxxResults_Manager.prototype.matchSuccess = function(numResults)
{
    log.debug("RxxResults_Manager.matchSuccess(numResults='"+numResults+"')");
    try {
        addStyle(gWidgets.resultsTitlebar, "background-color", "#E7E3DE");
        var label = "";
        if (gMode == "split") {
            label = "Split Results";
        } else if (gMode.slice(0, 5) == "match") {
            label = "Match Results";
        } else {
            label = "Replace Results";
        }
        if (gMode == "match") {
            label += " (match found)";
        } else {
            if (numResults == 1) {
                label += " (1 match found)";
            } else {
                label += " ("+numResults+" matches found)";
            }
        }
        gWidgets.resultsLabel.setAttribute("value", label);
        if (gMode == "split") {
            gWidgets.resultsDeck.selectedIndex = 2;
        } else {
            gWidgets.resultsDeck.selectedIndex = 1;
            if (gMode.slice(0, 5) == "match") {
                gWidgets.resultsReplacementColumn.setAttribute("collapsed", "true");
            } else {
                if (gWidgets.resultsReplacementColumn.hasAttribute("collapsed"))
                    gWidgets.resultsReplacementColumn.removeAttribute("collapsed");
            }
        }
    } catch(ex) {
        log.exception(ex);
    }
}

RxxResults_Manager.prototype.matchFailure = function(message)
{
    log.debug("RxxResults_Manager.matchFailure(message='"+message+"')");
    try {
        gWidgets.resultsLabel.setAttribute("value",
                                           gModeToName[gMode]+" Results");
        addStyle(gWidgets.resultsTitlebar, "background-color",
                 "#E7E3DE");
        gWidgets.resultsDeck.selectedIndex = 0;
        var msg = "Your regular expression does not match the search text";
        if (message) {
            msg += ": "+message;
        } else {
            msg += ".";
        }
        gWidgets.resultsMessage.setAttribute("value", msg);
        addStyle(gWidgets.resultsMessage, "color", "black");
    } catch(ex) {
        log.exception(ex);
    }
}

RxxResults_Manager.prototype.matchError = function(message)
{
    log.debug("RxxResults_Manager.matchError(message='"+message+"')");
    try {
        gWidgets.resultsLabel.setAttribute("value",
                                           gModeToName[gMode]+" Results");
        var red = "#b82619"; // fairly dark red, same as button-link:hover
        addStyle(gWidgets.resultsTitlebar, "background-color", red);
        gWidgets.resultsDeck.selectedIndex = 0;
        var msg;
        msg = "There is an error in your regular expression";
        if (message) {
            msg += ": "+message;
        } else {
            msg += ".";
        }
        gWidgets.resultsMessage.setAttribute("value", msg);
            //"There is an error in your regular expression: "+message);
        addStyle(gWidgets.resultsMessage, "color", red);
    } catch(ex) {
        log.exception(ex);
    }
}

RxxResults_Manager.prototype.replaceError = function(message)
{
    log.debug("RxxResults_Manager.replaceError(message='"+message+"')");
    try {
        gWidgets.resultsLabel.setAttribute("value",
                                           gModeToName[gMode]+" Results");
        var red = "#b82619"; // fairly dark red, same as button-link:hover
        addStyle(gWidgets.resultsTitlebar, "background-color", red);
        gWidgets.resultsDeck.selectedIndex = 0;
        var msg = "There is an error in your replacement template";
        if (message) {
            msg += ": "+message;
        } else {
            msg += ".";
        }
        gWidgets.resultsMessage.setAttribute("value", msg);
            //"There is an error in your regular expression: "+message);
        addStyle(gWidgets.resultsMessage, "color", red);
    } catch(ex) {
        log.exception(ex);
    }
}

RxxResults_Manager.prototype.lexError = function(message)
{
    log.debug("RxxResults_Manager.lexError(message='"+message+"')");
    try {
        if (gWidgets.resultsLexErrorMessage.hasAttribute("collapsed"))
            gWidgets.resultsLexErrorMessage.removeAttribute("collapsed");
        gWidgets.resultsLexErrorMessage.setAttribute("value", "("+message+")");
    } catch(ex) {
        log.exception(ex);
    }
}

RxxResults_Manager.prototype.lexSuccess = function()
{
    log.debug("RxxResults_Manager.lexSuccess()");
    try {
        gWidgets.resultsLexErrorMessage.setAttribute("collapsed", "true");
        gWidgets.resultsLexErrorMessage.setAttribute("value", "");
    } catch(ex) {
        log.exception(ex);
    }
}



//---- the scintilla/SciMoz eventists (aka event handlers)

function RxxRegex_SciMozEventist() {}
RxxRegex_SciMozEventist.prototype = {
    onStyleNeeded: function(position) { return false; },
    onCharAdded: function(ch) { return false; },
    onSavePointReached: function() { return false; },
    onSavePointLeft: function() { return false; },
    onModifyAttemptRO: function() { return false; },
    onKey: function(ch, modifiers) { return false; },
    onDoubleClick: function() { return false; },
    onUpdateUI: function() { return false; },
    onModified: function(position, modificationType, text, length, linesAdded,
                         line, foldLevelNow, foldLevelPrev)
    {
        try {
            if (modificationType & (SC_MOD_INSERTTEXT | SC_MOD_DELETETEXT)) {
                // Have to call in a timeout to avoid recusively calling
                // into scintilla (in which case the recursive calls are
                // silently ignored and no styling happens.)
                window.setTimeout("RxxRegex_Restyle();", 0);

                gNeedToRebuildReplacementShortcuts = true;
                RxxResults_Update();
                return true;
            }
        } catch(ex) {
            log.exception(ex);
        }
        return false;
    },
    onMacroRecord: function(message, wParam, lparam) {  return false; },
    onMarginClick: function(modifiers, position, margin) { return false; },
    onNeedShown: function(position, length) { return false; },
    onPosChanged: function(position) {
        //XXX Want to update some status text and/or perhaps a highlighted
        //    position in match results -- or something like that.
        //dump("XXX TODO: regex view pos changes: position="+position+"\n");
        return false;
    },
    // nsISupports interface
    QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsISupports) &&
            !iid.equals(Components.interfaces.ISciMozEvents))
        throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

function RxxReplacement_SciMozEventist() {}
RxxReplacement_SciMozEventist.prototype = {
    onStyleNeeded: function(position) { return false; },
    onCharAdded: function(ch) { return false; },
    onSavePointReached: function() { return false; },
    onSavePointLeft: function() { return false; },
    onModifyAttemptRO: function() { return false; },
    onKey: function(ch, modifiers) { return false; },
    onDoubleClick: function() { return false; },
    onUpdateUI: function() { return false; },
    onModified: function(position, modificationType, text, length, linesAdded,
                         line, foldLevelNow, foldLevelPrev)
    {
        try {
            if (modificationType & (SC_MOD_INSERTTEXT | SC_MOD_DELETETEXT)) {
                // Have to call in a timeout to avoid recusively calling
                // into scintilla (in which case the recursive calls are
                // silently ignored and no styling happens.)
                window.setTimeout("RxxReplacement_Restyle();", 0);

                RxxResults_Update();
                return true;
            }
        } catch(ex) {
            log.exception(ex);
        }
        return false;
    },
    onMacroRecord: function(message, wParam, lparam) {  return false; },
    onMarginClick: function(modifiers, position, margin) { return false; },
    onNeedShown: function(position, length) { return false; },
    onPosChanged: function(position) { return false; },
    // nsISupports interface
    QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsISupports) &&
            !iid.equals(Components.interfaces.ISciMozEvents))
        throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

function RxxSearchText_SciMozEventist() {}
RxxSearchText_SciMozEventist.prototype = {
    onStyleNeeded: function(position) { return false; },
    onCharAdded: function(ch) { return false; },
    onSavePointReached: function() { return false; },
    onSavePointLeft: function() { return false; },
    onModifyAttemptRO: function() { return false; },
    onKey: function(ch, modifiers) { return false; },
    onDoubleClick: function() { return false; },
    onUpdateUI: function() { return false; },
    onModified: function(position, modificationType, text, length, linesAdded,
                         line, foldLevelNow, foldLevelPrev)
    {
        try {
            if (modificationType & (SC_MOD_INSERTTEXT | SC_MOD_DELETETEXT)) {
                // Have to call in a timeout to avoid recusively calling
                // into scintilla (in which case the recursive calls are
                // silently ignored and no styling happens.)
                window.setTimeout("RxxResults_Update();", 0);
                return true;
            }
        } catch(ex) {
            log.exception(ex);
        }
        return false;
    },
    onMacroRecord: function(message, wParam, lparam) {  return false; },
    onMarginClick: function(modifiers, position, margin) { return false; },
    onNeedShown: function(position, length) { return false; },
    onPosChanged: function(position) { return false; },
    // nsISupports interface
    QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsISupports) &&
            !iid.equals(Components.interfaces.ISciMozEvents))
        throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

//---- interface routines for XUL

function OnLoad()
{
    log.debug("OnLoad()");
    try {
        scintillaOverlayOnLoad();

        gPrefs = Components.classes["@activestate.com/koPrefService;1"].
                 getService(Components.interfaces.koIPrefService).prefs;

        gWidgets = new Object();
        gWidgets.window = document.getElementById("komodo_rxx");
        gWidgets.buttonBox = document.getElementById("button-box");
        gWidgets.contentBox = document.getElementById("content-box");
        gWidgets.inputGrid = document.getElementById("input-grid");
        gWidgets.statusbarBox = document.getElementById("statusbar-box");
        //XXX No mode explanation while we only support one language.
        //gWidgets.modeExplanation = document.getElementById("mode-explanation");
        gWidgets.modeButtons = {
            "match": document.getElementById("match-mode-button"),
            "match-all": document.getElementById("match-all-mode-button"),
            "split": document.getElementById("split-mode-button"),
            "replace": document.getElementById("replace-mode-button"),
            "replace-all": document.getElementById("replace-all-mode-button")
        };
        gWidgets.searchColumn = document.getElementById("search-column");
        gWidgets.columnSplitter = document.getElementById("column-splitter");
        gWidgets.replaceColumn = document.getElementById("replace-column");
        gWidgets.patternRow = document.getElementById("pattern-row");
        gWidgets.rowSplitter = document.getElementById("row-splitter");
        gWidgets.textRow = document.getElementById("text-row");
        gWidgets.regexHeaderLabel = document.getElementById("regex-header-label");
        gWidgets.regexView = document.getElementById("regex-view");
        gWidgets.replacementView = document.getElementById("replacement-view");
        gWidgets.searchTextView = document.getElementById("search-text-view");
        gWidgets.replacedTextView = document.getElementById("replaced-text-view");
        gWidgets.regexArea = document.getElementById("regex-area");
        gWidgets.replacementArea = document.getElementById("replacement-area");
        gWidgets.searchTextArea = document.getElementById("search-text-area");
        gWidgets.replacedTextArea = document.getElementById("replaced-text-area");
        gWidgets.resultsTitlebar = document.getElementById("results-titlebar");
        gWidgets.resultsArea = document.getElementById("results-area");
        gWidgets.resultsLabel = document.getElementById("results-label");
        gWidgets.resultsDeck = document.getElementById("results-deck");
        gWidgets.resultsMessage = document.getElementById("results-message");
        gWidgets.resultsReplacementColumn = document.getElementById("match-results-tree-replacement");
        gWidgets.resultsLexErrorMessage = document.getElementById("results-lex-error-message");
        gWidgets.matchResultsTree = document.getElementById("match-results-tree");
        gWidgets.splitResultsTree = document.getElementById("split-results-tree");
        gWidgets.regexOptionICheckbox = document.getElementById("regex-option-i-checkbox");
        gWidgets.regexOptionMCheckbox = document.getElementById("regex-option-m-checkbox");
        gWidgets.regexOptionSCheckbox = document.getElementById("regex-option-s-checkbox");
        gWidgets.regexOptionXCheckbox = document.getElementById("regex-option-x-checkbox");
        gWidgets.regexOptionUCheckbox = document.getElementById("regex-option-u-checkbox");
        gWidgets.regexOptionLCheckbox = document.getElementById("regex-option-l-checkbox");

        gWidgets.regexView.initWithBuffer("", "Regex");
        var scimoz = gWidgets.regexView.scimoz;
        scimoz.viewWS = 1;
        scimoz.visible = 1;
        scimoz.viewEOL = 1;
        scimoz.caretLineVisible = 0;
        gRegexEventist = new RxxRegex_SciMozEventist();
        scimoz.hookEvents(gRegexEventist);
        scimoz.lexer = scimoz.SCLEX_CONTAINER;
        scimoz.setCommandUpdateTarget(window);
        // Use the first indicator to mark regex syntax errors.
        scimoz.indicSetStyle(0, scimoz.INDIC_SQUIGGLE);
        scimoz.indicSetFore(0, 0x0000ff);

        gWidgets.replacementView.initWithBuffer("", "Regex");
        scimoz = gWidgets.replacementView.scimoz;
        scimoz.viewWS = 1;
        scimoz.visible = 1;
        scimoz.caretLineVisible = 0;
        gReplacementEventist = new RxxReplacement_SciMozEventist();
        scimoz.hookEvents(gReplacementEventist);
        scimoz.lexer = scimoz.SCLEX_CONTAINER;
        scimoz.setCommandUpdateTarget(window);
        // Use the first indicator to mark replacement syntax errors.
        scimoz.indicSetStyle(0, scimoz.INDIC_SQUIGGLE);
        scimoz.indicSetFore(0, 0x0000ff);

        gWidgets.searchTextView.initWithBuffer("", "Regex");
        scimoz = gWidgets.searchTextView.scimoz;
        scimoz.viewWS = 1;
        scimoz.visible = 1;
        scimoz.caretLineVisible = 0;
        gSearchTextEventist = new RxxSearchText_SciMozEventist();
        scimoz.hookEvents(gSearchTextEventist);
        scimoz.lexer = scimoz.SCLEX_CONTAINER;
        scimoz.setCommandUpdateTarget(window);

        gWidgets.replacedTextView.initWithBuffer("", "Regex");
        scimoz = gWidgets.replacedTextView.scimoz;
        scimoz.readOnly = 1;
        scimoz.viewWS = 1;
        scimoz.visible = 1;
        scimoz.caretLineVisible = 0;
        // Don't care about events in this scimoz, so don't bother with
        // an eventist.
        //gReplacedTextEventist = new RxxReplacedText_SciMozEventist();
        //scimoz.hookEvents(gReplacedTextEventist);
        scimoz.lexer = scimoz.SCLEX_CONTAINER;
        scimoz.setCommandUpdateTarget(window);

        // Setup the match results pane.
        gResultsManager = new RxxResults_Manager();

        // .language
        var language = null;
        if (typeof(window.arguments[0].language) == "undefined"
            || window.arguments[0].language == null) {
            language = gPrefs.getStringPref("rxx_language");
        } else {
            language = window.arguments[0].language;
        }
        Rxx_ChangeLanguage(language, true);

        // .<some attribute(s) for regex options>
        //XXX Do we need window input arguments for the regex options?
        gRegexOptions = Components.classes["@activestate.com/koRxxRegexOptions;1"]
                        .createInstance(Components.interfaces.koIRxxRegexOptions);
        gRegexOptions.restoreFromPrefs();
        gWidgets.regexOptionICheckbox.checked = gRegexOptions.i;
        gWidgets.regexOptionMCheckbox.checked = gRegexOptions.m;
        gWidgets.regexOptionSCheckbox.checked = gRegexOptions.s;
        gWidgets.regexOptionXCheckbox.checked = gRegexOptions.x;
        gWidgets.regexOptionUCheckbox.checked = gRegexOptions.u;
        gWidgets.regexOptionLCheckbox.checked = gRegexOptions.l;

        // .mode
        var mode = null;
        if (typeof(window.arguments[0].mode) == "undefined"
            || window.arguments[0].mode == null) {
            mode = gPrefs.getStringPref("rxx_mode");
        } else {
            mode = window.arguments[0].mode;
        }
        if (!mode) mode = "match-all";
        Rxx_SetMode(mode, true);
        gWidgets.modeButtons[gMode].setAttribute("checked", "true");

        // .regex
        var regex = null;
        if (typeof(window.arguments[0].regex) == "undefined"
            || window.arguments[0].regex == null)
        {
            // Use the current selection, if there is one, for the regex.
            //XXX Longer run should attempt to parse out a "regex under
            //    cursor if the style looks right.
            var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                               .getService(Components.interfaces.nsIWindowMediator);
            var w = wm.getMostRecentWindow("Komodo");
            if (w == opener) {
                var v = opener.gViewMgr.currentView;
                if (v && v.scimoz) {
                    regex = v.scimoz.selText;
                }
            }

            if (regex) {
                // do nothing, already retrieved from current selection
            } else if (gPrefs.hasStringPref("rxx_regex")) {
                regex = gPrefs.getStringPref("rxx_regex");
            } else if (typeof(gSampleRegex[gLanguage]) != "undefined") {
                regex = gSampleRegex[gLanguage];
            } else {
                regex = gSampleRegex["*"];
            }
        } else {
            regex = window.arguments[0].regex;
        }
        gWidgets.regexView.setBufferText(regex);
        // Do NOT select the regex on start up because (1) when clicking
        // around in the UI the selection of the regex string is just a
        // little bit more confusion and (2) on Linux, clicking in the
        // selection area gets HALF way to getting full cursor focus to that
        // scintilla. Without a selection, clicking in works better.
        gWidgets.regexView.scimoz.anchor = regex.length;
        gWidgets.regexView.scimoz.currentPos = regex.length;

        // .replacement
        var replacement = null;
        if (typeof(window.arguments[0].replacement) == "undefined"
            || window.arguments[0].replacement == null)
        {
            if (gPrefs.hasStringPref("rxx_replacement")) {
                replacement = gPrefs.getStringPref("rxx_replacement");
            } else if (typeof(gSampleReplacement[gLanguage]) != "undefined") {
                replacement = gSampleReplacement[gLanguage];
            } else {
                replacement = gSampleReplacement["*"];
            }
        } else {
            replacement = window.arguments[0].replacement;
        }
        gWidgets.replacementView.setBufferText(replacement);
        //gWidgets.replacementView.scimoz.anchor = 0;
        //gWidgets.replacementView.scimoz.currentPos = replacement.length;

        // .searchText
        var searchText = null;
        if (typeof(window.arguments[0].searchText) == "undefined"
            || window.arguments[0].searchText == null)
        {
            if (gPrefs.hasStringPref("rxx_search_text")) {
                searchText = gPrefs.getStringPref("rxx_search_text");
            } else if (typeof(gSampleSearchText[gLanguage]) != "undefined") {
                searchText = gSampleSearchText[gLanguage];
            } else {
                searchText = gSampleSearchText["*"];
            }
        } else {
            searchText = window.arguments[0].searchText;
        }
        gWidgets.searchTextView.setBufferText(searchText);

        RxxResults_Update();

        // Restore the window dimensions.
        var height, width;
        if (gPrefs.hasLongPref("rxx_pattern_row_height")) {
            height = gPrefs.getLongPref("rxx_pattern_row_height");
            if (height) {
                gWidgets.patternRow.setAttribute("height", height);
            }
        }
        if (gPrefs.hasLongPref("rxx_text_row_height")) {
            height = gPrefs.getLongPref("rxx_text_row_height");
            if (height) {
                gWidgets.textRow.setAttribute("height", height);
            }
        }
        if (gPrefs.hasLongPref("rxx_search_column_width")) {
            width = gPrefs.getLongPref("rxx_search_column_width");
            if (width) {
                gWidgets.searchColumn.setAttribute("width", width);
            }
        }
        if (gPrefs.hasLongPref("rxx_replace_column_width")) {
            width = gPrefs.getLongPref("rxx_replace_column_width");
            if (width) {
                gWidgets.replaceColumn.setAttribute("width", width);
            }
        }
        if (gPrefs.hasLongPref("rxx_match_results_area_height")) {
            height = gPrefs.getLongPref("rxx_match_results_area_height");
            if (height) {
                gWidgets.resultsArea.setAttribute("height", height);
            }
        }

        // Focus the regex view.
        //XXX If I do not put this in a timeout the window gets drawn at
        //    0,0 and then is quickly moved to the persisted position. The
        //    flashing from this is very annoying.
        window.setTimeout("gWidgets.regexView.setFocus();", 0);
    } catch(ex) {
        log.exception(ex);
    }
}


function OnUnload()
{
    log.debug("OnUnload()");
    try {
        gPrefs.setLongPref("rxx_pattern_row_height",
            gWidgets.patternRow.boxObject.height);
        gPrefs.setLongPref("rxx_text_row_height",
            gWidgets.textRow.boxObject.height);
        gPrefs.setLongPref("rxx_search_column_width",
            gWidgets.searchColumn.boxObject.width);
        gPrefs.setLongPref("rxx_replace_column_width",
            gWidgets.replaceColumn.boxObject.width);
        gPrefs.setLongPref("rxx_match_results_area_height",
            gWidgets.resultsArea.boxObject.height);
        gPrefs.setStringPref("rxx_mode", gMode);
        gPrefs.setStringPref("rxx_language", gLanguage);
        gPrefs.setStringPref("rxx_regex", gWidgets.regexView.scimoz.text);
        gPrefs.setStringPref("rxx_replacement", gWidgets.replacementView.scimoz.text);
        gPrefs.setStringPref("rxx_search_text", gWidgets.searchTextView.scimoz.text);
        gRegexOptions.saveToPrefs();

        gWidgets.regexView.scimoz.unhookEvents(gRegexEventist);
        gRegexEventist = null;
        gWidgets.replacementView.scimoz.unhookEvents(gReplacementEventist);
        gReplacementEventist = null;
        gWidgets.searchTextView.scimoz.unhookEvents(gSearchTextEventist);
        gSearchTextEventist = null;
    } catch(ex) {
        log.exception(ex);
    }
}


function Rxx_DumpDimensions() // for debugging
{
    try {
        dump("----------------- Rxx Window Dimensions ----------------\n");
        dump("window height: "+window.height+" (inner height:"+
             window.innerHeight+", outer height:"+window.outerHeight+")\n");
        dump("    buttonBox height:"+
             gWidgets.buttonBox.getAttribute("height")+
             " ("+gWidgets.buttonBox.boxObject.height+")\n");
        dump("    contentBox height:"+
             gWidgets.contentBox.getAttribute("height")+
             " ("+gWidgets.contentBox.boxObject.height+")\n");
        dump("        inputGrid height:"+
             gWidgets.inputGrid.getAttribute("height")+
             " ("+gWidgets.inputGrid.boxObject.height+")\n");
        dump("            patternRow height:"+
             gWidgets.patternRow.getAttribute("height")+
             " ("+gWidgets.patternRow.boxObject.height+")\n");
        dump("            textRow height:"+
             gWidgets.textRow.getAttribute("height")+
             " ("+gWidgets.textRow.boxObject.height+")\n");
        dump("        resultsArea height:"+
             gWidgets.resultsArea.getAttribute("height")+
             " ("+gWidgets.resultsArea.boxObject.height+")\n");
        dump("    statusbarBox height:"+
             gWidgets.statusbarBox.getAttribute("height")+
             " ("+gWidgets.statusbarBox.boxObject.height+")\n");
        dump("--------------------------------------------------------\n");
    } catch(ex) {
        log.exception(ex);
    }
}


function _Rxx_UpdateModeExplanation(language, mode)
{
    var explanation = "";

    var langMode = gLanguage+"-"+mode;
    switch(langMode) {
    case "python-match":
        explanation = "re.search(...)";
        break;
    case "python-match-all":
        explanation = "re.findall(...)";
        break;
    case "python-split":
        explanation = "re.split(...)";
        break;
    case "python-replace":
        explanation = "re.sub(..., 1)";
        break;
    case "python-replace-all":
        explanation = "re.sub(...)";
        break;
    case "perl-match":
        explanation = "m/.../";
        break;
    case "perl-match-all":
        explanation = "m/.../g";
        break;
    case "perl-split":
        explanation = "split /.../";
        break;
    case "perl-replace":
        explanation = "s/.../";
        break;
    case "perl-replace-all":
        explanation = "s/.../g";
        break;
    default:
        var msg = "Unexpected language and mode: '"+langMode+"'.";
        dialog_internalError(msg, msg);
    }
    //XXX No mode explanation while we only have one language.
    //gWidgets.modeExplanation.setAttribute("value", explanation);
}

// Showing/Hiding the "replace column".
// For now we just (un)collapse this column. Later we may want to do fancy
// animation of the column showing/hiding and possibly having it
// expand/contract the window width as necessary to keep the search column
// width constanct.
function _Rxx_ShowReplaceColumn()
{
    if (gWidgets.columnSplitter.hasAttribute("collapsed"))
        gWidgets.columnSplitter.removeAttribute("collapsed");
    if (gWidgets.replaceColumn.hasAttribute("collapsed"))
        gWidgets.replaceColumn.removeAttribute("collapsed");
    if (gWidgets.replacedTextView.hasAttribute("collapsed"))
        gWidgets.replacedTextView.removeAttribute("collapsed");
    if (gWidgets.replacementView.hasAttribute("collapsed"))
        gWidgets.replacementView.removeAttribute("collapsed");
}

function _Rxx_HideReplaceColumn()
{
    gWidgets.columnSplitter.setAttribute("collapsed", "true");
    gWidgets.replaceColumn.setAttribute("collapsed", "true");
    // Need to specifically collapsed the scintilla plugins, otherwise a
    // thin (couple pixel) strip of them shows up on the right edge.
    gWidgets.replacedTextView.setAttribute("collapsed", "true");
    gWidgets.replacementView.setAttribute("collapsed", "true");
}


function Rxx_SetMode(mode, initializing /* = false */) {
    if (typeof(initializing) == "undefined" || initializing == null) initializing = false;
    log.debug("Rxx_SetMode(mode='"+mode+"', initializing="+initializing+")");

    try {
        gMode = mode;
        _Rxx_UpdateModeExplanation(gLanguage, gMode);
        if (gMode.slice(0, 7) == "replace") {
            _Rxx_ShowReplaceColumn();
        } else {
            _Rxx_HideReplaceColumn();
        }
        if (!initializing) {
            RxxResults_Update();
        }
    } catch(ex) {
        log.exception(ex);
    }
}


function Rxx_LoadSample() {
    log.debug("Rxx_LoadSample()");
    try {
        var regex;
        if (typeof(gSampleRegex[gLanguage]) != "undefined") {
            regex = gSampleRegex[gLanguage];
        } else {
            regex = gSampleRegex["*"];
        }
        var searchText;
        if (typeof(gSampleSearchText[gLanguage]) != "undefined") {
            searchText = gSampleSearchText[gLanguage];
        } else {
            searchText = gSampleSearchText["*"];
        }
        var replacement;
        if (typeof(gSampleReplacement[gLanguage]) != "undefined") {
            replacement = gSampleReplacement[gLanguage];
        } else {
            replacement = gSampleReplacement["*"];
        }

        gWidgets.regexView.setBufferText(regex);
        gWidgets.searchTextView.setBufferText(searchText);
        gWidgets.replacementView.setBufferText(replacement);
    } catch(ex) {
        log.exception(ex);
    }
}


function RxxRegex_Restyle() {
    log.debug("RxxRegex_Restyle()");
    try {
        gHandler.styleRegex(gWidgets.regexView.scimoz, gRegexOptions,
                            gResultsManager);
    } catch(ex) {
        log.exception(ex);
    }
}

function RxxReplacement_Restyle() {
    log.debug("RxxReplacement_Restyle()");
    try {
        gHandler.styleReplacement(gWidgets.replacementView.scimoz,
                                  gResultsManager);
    } catch(ex) {
        log.exception(ex);
    }
}


// Open the given help category.
function Rxx_MainHelp(page)
{
    log.debug("Rxx_MainHelp(page='"+page+"')");
    try {
        gMainWindow.launch_MainHelp(page);
    } catch (ex) {
        log.exception(ex);
    }
}
function Rxx_WebHelp(url)
{
    log.debug("Rxx_WebHelp(url='"+url+"')");
    try {
        gMainWindow.browse_OpenUrlInDefaultBrowser(url);
    } catch (ex) {
        log.exception(ex);
    }
}


function Rxx_ChangeLanguage(language, initializing /* = false */)
{
    log.debug("Rxx_ChangeLanguage(language='"+language+"')");
    try {
        if (typeof(initializing) == "undefined" || initializing == null) initializing = false;

        if (! gLanguageToDisplayName[language]) {
            log.warn("bogus regex language '"+language+
                     "', falling back to 'python'");
            language = "python";
        }

        gLanguage = language;
        // Disabling updating of header until we can switch languages.
        //gWidgets.regexHeaderLabel.setAttribute("value",
        //    gLanguageToDisplayName[gLanguage]+" Regular Expression");
        if (gHandler == null || gHandler.language != gLanguage) {
            var handlerCID = gLanguageToHandlerCID[gLanguage];
            if (handlerCID == null) {
                var err = "There is no Rxx handler for the '"+
                          gLanguage+"' regex language!";
                dialog_internalError(err, err, opener);
                gHandler = null;
            } else {
                gHandler = Components.classes[handlerCID]
                    .createInstance(Components.interfaces.koIRxxHandler);
            }
        }
        gNeedToRebuildRegexShortcuts = true;
        gNeedToRebuildReplacementShortcuts = true;

        if (!initializing) {
            _Rxx_UpdateModeExplanation(gLanguage, gMode);
            RxxResults_Update();
        }
    } catch (ex) {
        log.exception(ex);
    }
}


function RxxRegex_UpdateOption(option)
{
    log.debug("RxxRegex_UpdateOption()");
    try {
        switch (option) {
        case 'i':
            gRegexOptions.i = gWidgets.regexOptionICheckbox.checked;
            break;
        case 'm':
            gRegexOptions.m = gWidgets.regexOptionMCheckbox.checked;
            break;
        case 's':
            gRegexOptions.s = gWidgets.regexOptionSCheckbox.checked;
            break;
        case 'x':
            gRegexOptions.x = gWidgets.regexOptionXCheckbox.checked;
            break;
        case 'u':
            gRegexOptions.u = gWidgets.regexOptionUCheckbox.checked;
            break;
        case 'l':
            gRegexOptions.l = gWidgets.regexOptionLCheckbox.checked;
            break;
        default:
            throw("unexpected regex option: '"+option+"'\n");
        }
        RxxResults_Update();
    } catch (ex) {
        log.exception(ex);
    }
}


// Build the regex syntax shortcuts menupopup, if necessary.
function RxxRegex_ShortcutsShowing(menupopup)
{
    log.debug("RxxRegex_ShortcutsShowing()");
    try {
        if (gNeedToRebuildRegexShortcuts) {
            var shortcuts = gRegexShortcuts[gLanguage];

            // Remove existing menuitems.
            while (menupopup.firstChild) {
                menupopup.removeChild(menupopup.firstChild);
            }

            var shortcut, mi, label;
            for (var i = 0; i < shortcuts.length; i++) {
                // Build a menuitem like this:
                //  <menuitem label="^ : beginning of line"
                //            shortcut="^"
                //            oncommand="RxxRegex_InsertShortcut(this);"/>
                // or this:
                //  <menuseparator/>
                shortcut = shortcuts[i];
                if (typeof(shortcut["separator"]) != "undefined") {
                    mi = document.createElementNS(XUL_NS, "menuseparator");
                } else {
                    mi = document.createElementNS(XUL_NS, "menuitem");
                    label = shortcut["shortcut"] + " : " + shortcut["desc"];
                    mi.setAttribute("label", label);
                    mi.setAttribute("shortcut", shortcut["shortcut"]);
                    mi.setAttribute("crop", "center");
                    mi.setAttribute("oncommand", "RxxRegex_InsertShortcut(this);");
                    if (typeof(shortcut["tooltip"]) != "undefined") {
                        mi.setAttribute("tooltiptext", shortcut["tooltip"]);
                    }
                }
                menupopup.appendChild(mi);
            }
            gNeedToRebuildRegexShortcuts = false;
        }
    } catch (ex) {
        log.exception(ex);
    }
}


// Build the replacement syntax shortcuts menupopup, if necessary.
function RxxReplacement_ShortcutsShowing(menupopup)
{
    log.debug("RxxReplacement_ShortcutsShowing()");
    try {
        if (gNeedToRebuildReplacementShortcuts) {
            // Get group information for the current regex.
            var numGroups = gHandler.getLastNumGroups();
            var groupNamesObj = new Object();
            var countObj = new Object();
            gHandler.getLastGroupNames(countObj, groupNamesObj);
            var groupNames = groupNamesObj.value;

            // Remove existing menuitems.
            while (menupopup.firstChild) {
                menupopup.removeChild(menupopup.firstChild);
            }

            if (numGroups == 0) {
                // Build a menuitem like this:
                //  <menuitem label="(no groups in regex)"
                //            disabled="true"/>
                mi = document.createElementNS(XUL_NS, "menuitem");
                mi.setAttribute("label", "(no groups in regex)");
                mi.setAttribute("disabled", "true");
                menupopup.appendChild(mi);
            } else {
                var i, mi, label, name;
                for (i = 1; i < numGroups+1; i++) {
                    // Build a menuitem like this:
                    //  <menuitem label="group 0: \0"
                    //            shortcut="\0"
                    //            oncommand="RxxReplacement_InsertShortcut(this);"/>
                    mi = document.createElementNS(XUL_NS, "menuitem");
                    label = "group "+i+": \\"+i;
                    mi.setAttribute("label", label);
                    mi.setAttribute("shortcut", "\\"+i);
                    mi.setAttribute("oncommand", "RxxReplacement_InsertShortcut(this);");
                    menupopup.appendChild(mi);
                }
                if (groupNames.length > 0) {
                    mi = document.createElementNS(XUL_NS, "menuseparator");
                    menupopup.appendChild(mi);
                }
                for (i = 0; i < groupNames.length; i++) {
                    // Build a menuitem like this:
                    //  <menuitem label="group 'name': \g<name>"
                    //            shortcut="\g<name>"
                    //            oncommand="RxxReplacement_InsertShortcut(this);"/>
                    name = groupNames[i];
                    mi = document.createElementNS(XUL_NS, "menuitem");
                    label = "group '"+name+"': \\g<"+name+">";
                    mi.setAttribute("label", label);
                    mi.setAttribute("shortcut", "\\g<"+name+">");
                    mi.setAttribute("oncommand", "RxxReplacement_InsertShortcut(this);");
                    menupopup.appendChild(mi);
                }
            }
            gNeedToRebuildReplacementShortcuts = false;
        }
    } catch (ex) {
        log.exception(ex);
    }
}


// Insert the given "shortcut" into the given "scimoz" and select the
// inserted text.
function _insert(scimoz, shortcut)
{
    if (scimoz.anchor <= scimoz.currentPos) {
        scimoz.targetStart = scimoz.anchor;
        scimoz.targetEnd = scimoz.currentPos;
        scimoz.replaceTarget(shortcut.length, shortcut);
        scimoz.currentPos = scimoz.targetStart +
                            stringutils_bytelength(shortcut);
        if (scimoz.anchor == scimoz.currentPos) {
            scimoz.anchor = scimoz.targetStart +
                            stringutils_bytelength(shortcut);
        } else {
            scimoz.anchor = scimoz.targetStart;
        }
    } else {
        scimoz.targetStart = scimoz.currentPos;
        scimoz.targetEnd = scimoz.anchor;
        scimoz.replaceTarget(shortcut.length, shortcut);
        scimoz.currentPos = scimoz.targetStart;
        scimoz.anchor = scimoz.targetStart +
                        stringutils_bytelength(shortcut);
    }
}

function _Rxx_InsertShortcut(scimoz, shortcutWidget)
{
    //XXX:TODO: look at using snippets for inserting shortcuts, then can
    //          control selection bndries and cursor position and can use
    //          interpolation mechanisms. Would need to factor out the
    //          actual snippet insertion smarts from
    //          peSnippet.js::_insertSnippet() into a separate
    //          snippet_functions.js module.
    var shortcut = shortcutWidget.getAttribute("shortcut");
    var ellipsisIndex = shortcut.indexOf("...");

    // For bounding shortcuts (e.g., '(...)' is a bounding shortcut):
    // if there is a selection put it in as the '...' part, otherwise
    // insert and select the '...'.
    if (ellipsisIndex != -1) {
        var selection = scimoz.selText;
        if (selection) {
            shortcut = shortcut.replace(/\.\.\./, selection);
            _insert(scimoz, shortcut);
        } else {
            _insert(scimoz, shortcut);
            if (scimoz.anchor <= scimoz.currentPos) {
                scimoz.anchor += ellipsisIndex;
                scimoz.currentPos = scimoz.anchor + 3;
            } else {
                scimoz.anchor = scimoz.currentPos + ellipsisIndex;
                scimoz.currentPos = scimoz.anchor + 3;
            }
        }
    }
    // For non-bounding shortcuts (e.g., '^' is a non-bounding shortcut)
    // replace the selection if there is one or just insert at the
    // current pos.
    else {
        _insert(scimoz, shortcut);
    }
}

function RxxRegex_InsertShortcut(shortcutWidget)
{
    log.debug("RxxRegex_InsertShortcut()");
    try {
        _Rxx_InsertShortcut(gWidgets.regexView.scimoz, shortcutWidget);
        gWidgets.regexView.setFocus();
    } catch (ex) {
        log.exception(ex);
    }
}

function RxxReplacement_InsertShortcut(shortcutWidget)
{
    log.debug("RxxReplacement_InsertShortcut()");
    try {
        _Rxx_InsertShortcut(gWidgets.replacementView.scimoz, shortcutWidget);
        gWidgets.replacementView.setFocus();
    } catch (ex) {
        log.exception(ex);
    }
}


function RxxRegex_OnClick(event)
{
    log.debug("RxxRegex_OnClick(event)");
    //gMainWindow.gTimeline.dumpEvent(event);
    try {
        // This seems to be the signature resulting (strangely) from
        // invoking the accesskey for this view.
        if (event.detail == 0 && event.originalTarget.nodeName == "view") {
            gWidgets.regexView.setFocus();
            return false;
        }
    } catch (ex) {
        log.exception(ex);
    }
    return true;
}


function RxxReplacement_OnClick(event)
{
    log.debug("RxxReplacement_OnClick(event)");
    //gMainWindow.gTimeline.dumpEvent(event);
    try {
        // This seems to be the signature resulting (strangely) from
        // invoking the accesskey for this view.
        if (event.detail == 0 && event.originalTarget.nodeName == "view") {
            gWidgets.replacementView.setFocus();
            return false;
        }
    } catch (ex) {
        log.exception(ex);
    }
    return true;
}


function RxxSearchText_Open()
{
    log.debug("RxxSearchText_Open()");
    try {
        //XXX Should perhaps default to the directory of document of the view
        //    to which this "Rx" instance is attached.
        var filename = filepicker_openFile();
        if (filename != null) {
            var docSvc = Components.classes['@activestate.com/koDocumentService;1']
                        .getService(Components.interfaces.koIDocumentService);
            //XXX uriparse, docSvc, creating a koIDocument instance... it all
            //    seems overkill to just get the files contents.
            var uri = uriparse_pathToURI(filename);
            uri = uriparse_fixupURI(uri);
            var doc = docSvc.createDocumentFromURI(uri);
            doc.load();
            gWidgets.searchTextView.setBufferText(doc.buffer);

            RxxResults_Update();
        }
    } catch (ex) {
        log.exception(ex);
    }
}


function RxxSearchText_JumpToCurrentMatchResult()
{
    log.debug("RxxSearchText_JumpToCurrentMatchResult()");
    try {
        var start = gResultsManager.matchResultsView.currentMatchStart;
        var end = gResultsManager.matchResultsView.currentMatchEnd;
        var scimoz = gWidgets.searchTextView.scimoz;
        scimoz.setSel(start, end); // set selection and make it visible
        scimoz.chooseCaretX();
        gWidgets.searchTextView.setFocus();
    } catch (ex) {
        log.exception(ex);
    }
}


//XXX Could merge this with above and have the caller get the start/end.
function RxxSearchText_JumpToCurrentSplitResult()
{
    log.debug("RxxSearchText_JumpToCurrentSplitResult()");
    try {
        var start = gResultsManager.splitResultsView.currentResultStart;
        var end = gResultsManager.splitResultsView.currentResultEnd;
        var scimoz = gWidgets.searchTextView.scimoz;
        scimoz.setSel(start, end); // set selection and make it visible
        scimoz.chooseCaretX();
        gWidgets.searchTextView.setFocus();
    } catch (ex) {
        log.exception(ex);
    }
}


function RxxSearchText_OnClick(event)
{
    log.debug("RxxSearchText_OnClick(event)");
    //gMainWindow.gTimeline.dumpEvent(event);
    try {
        // This seems to be the signature resulting (strangely) from
        // invoking the accesskey for this view.
        if (event.detail == 0 && event.originalTarget.nodeName == "view") {
            gWidgets.searchTextView.setFocus();
            return false;
        }
    } catch (ex) {
        log.exception(ex);
    }
    return true;
}


function RxxReplacedText_OnClick(event)
{
    log.debug("RxxReplacedText_OnClick(event)");
    //gMainWindow.gTimeline.dumpEvent(event);
    try {
        // This seems to be the signature resulting (strangely) from
        // invoking the accesskey for this view.
        if (event.detail == 0 && event.originalTarget.nodeName == "view") {
            gWidgets.replacedTextView.setFocus();
            return false;
        }
    } catch (ex) {
        log.exception(ex);
    }
    return true;
}


// Attempt to match the regex to the search text and display any results.
function RxxResults_Update()
{
    log.debug("RxxResults_Update()");
    try {
        var pattern = gWidgets.regexView.scimoz.text;
        var replacement;
        if (pattern) {
            switch (gMode) {
            case "match":
                gHandler.match(pattern, gRegexOptions,
                               gWidgets.searchTextView.scimoz,
                               gResultsManager);
                break;
            case "match-all":
                gHandler.matchAll(pattern, gRegexOptions,
                                  gWidgets.searchTextView.scimoz,
                                  gResultsManager);
                break;
            case "split":
                gHandler.split(pattern, gRegexOptions,
                               gWidgets.searchTextView.scimoz,
                               gResultsManager);
                break;
            case "replace":
                replacement = gWidgets.replacementView.scimoz.text;
                gHandler.replace(pattern, replacement, gRegexOptions,
                                 gWidgets.searchTextView.scimoz,
                                 gWidgets.replacedTextView.scimoz,
                                 gResultsManager);
                break;
            case "replace-all":
                replacement = gWidgets.replacementView.scimoz.text;
                gHandler.replaceAll(pattern, replacement, gRegexOptions,
                                    gWidgets.searchTextView.scimoz,
                                    gWidgets.replacedTextView.scimoz,
                                    gResultsManager);
                break;
            }
            gResultsManager.matchResultsView.selection.clearSelection();
        } else {
            gWidgets.resultsLabel.setAttribute("value", "Match Results");
            addStyle(gWidgets.resultsTitlebar, "background-color",
                     "#E7E3DE");
            gWidgets.resultsDeck.selectedIndex = 0;
            var msg = "The regular expression is empty.";
            gWidgets.resultsMessage.setAttribute("value", msg);
            addStyle(gWidgets.resultsMessage, "color", "black");
        }
    } catch (ex) {
        log.exception(ex);
    }
}


function RxxResults_OnClick(event)
{
    log.debug("RxxResults_OnClick(event)");
    //gMainWindow.gTimeline.dumpEvent(event);
    try {
        // c.f. mozilla/mailnews/base/resources/content/threadPane.js
        var t = event.originalTarget;

        // single-click on a column
        if (t.localName == "treecol") {
            //XXX Ignore column clicks for now, ordering with hierarchy
            //    is awkward.
            // t.id is the id of the clicked column
        }

        // unmodified left double-click in the tree body
        else if (event.detail == 2 && event.button == 0 &&
                 t.localName == "treechildren" &&
                 event.shiftKey == false && event.ctrlKey == false &&
                 event.altKey == false && event.metaKey == false) {
            if (gMode == "split") {
                RxxSearchText_JumpToCurrentSplitResult();
            } else {
                RxxSearchText_JumpToCurrentMatchResult();
            }
            //XXX The following aren't working. Grrr. Is that because this
            //    fires on the wrong phase?
            event.cancelBubble = true;
            return false;
        }

        // This seems to be the signature resulting (strangely) from
        // invoking the accesskey for the treechildren widget. Make sure
        // the fact that the Match Results are focussed is visible.
        else if (event.detail == 0 && t.localName == "deck") {
            var view = null;
            var tree = null;
            if (gMode == "split") {
                view = gResultsManager.splitResultsView;
                tree = gWidgets.splitResultsTree;
            } else {
                view = gResultsManager.matchResultsView;
                tree = gWidgets.matchResultsTree;
            }
            var currentIndex = view.selection.currentIndex;
            if (currentIndex == -1) {
                currentIndex = 0;
            } else if (currentIndex >= view.rowCount) {
                currentIndex = view.rowCount - 1;
            }
            view.selection.select(currentIndex);
            tree.focus();
        }
    } catch (ex) {
        log.exception(ex);
    }
    return true;
}


function RxxResults_OnKeyPress(event)
{
    log.debug("RxxResults_OnKeyPress(event)");
    //gMainWindow.gTimeline.dumpEvent(event);
    try {
        if (event.keyCode == 13 /* <Enter> */ && gMode != "split") {
            RxxSearchText_JumpToCurrentMatchResult();
            //XXX The following aren't working. Grrr. Is that because this
            //    fires on the wrong phase?
            event.cancelBubble = true;
            return false;
        }
    } catch (ex) {
        log.exception(ex);
    }
    return true;
}


