#!python
# Copyright (c) 2000-2003 ActiveState Corp.
# See the file LICENSE.txt for licensing information.

# A service for providing Komodo-specific directory/filename information.

from xpcom import components
import sys, os, re
import logging


log = logging.getLogger("koDirs")

#---- internal support stuff

def SmallestVersionFirst(a, b):
    """Compare two version strings and return:
        -1 if a < b
         0 if a == b
        +1 if a > b
    """
    return cmp([int(elem) for elem in a.split('.')],
               [int(elem) for elem in b.split('.')])


#---- component implementation

class KoDirs:
    _com_interfaces_ = [components.interfaces.koIDirs]
    _reg_clsid_ = "{98435d6d-e24d-4057-a72a-dfcd1d282fae}"
    _reg_contractid_ = "@activestate.com/koDirs;1"
    _reg_desc_ = "Komodo Directory Information Service"

    def __init__(self):
        pass

    def _userAppDataPath(self):
        # Allow a special environment variable to override the User Data
        # Dir for a Komodo run. The main motivation for this is bug
        # 32270.
        envPath = os.environ.get("KOMODO_USERDATADIR")
        if envPath:
            path = os.path.expanduser(envPath)
        elif sys.platform.startswith("win"):
            # try to make this a unicode path bug #26404
            # SHGetFolderPath does not return unicode strings when
            # there is unicode data in the path
            from win32com.shell import shellcon, shell
            path = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
            try:
                path = unicode(path)
            except:
                pass
            path = os.path.join(path, "ActiveState", "Komodo")
        else:
            path = os.path.expanduser("~/.komodo")
        return path

    def _commonAppDataPath(self):
        if sys.platform.startswith("win"):
            # try to make this a unicode path bug #26404
            # SHGetFolderPath does not return unicode strings when
            # there is unicode data in the path
            from win32com.shell import shellcon, shell
            shellcon.CSIDL_COMMON_APPDATA = 0x23 # missing from shellcon
            path = shell.SHGetFolderPath(0, shellcon.CSIDL_COMMON_APPDATA, 0, 0)
            try:
                path = unicode(path)
            except:
                pass
            path = os.path.join(path, "ActiveState", "Komodo")
        else:
            path = "/etc/komodo"
        return path

    def GetAllUserDataDirs(self):
        """Return list of the full paths to all existing Komodo
        user data dirs. This list is ordered from lowest Komodo version
        to highest."""
        base = self._userAppDataPath()
        try:
            paths = os.listdir(base)
        except OSError:
            # 'base' does not exist
            return []
        userDataDirDict = {}
        versionRe = re.compile("^\d+\.\d+$")
        for path in paths:
            fullPath = os.path.join(base, path)
            if os.path.isdir(fullPath):
                # Ensure it is a Komodo version dir (i.e. noone came in an
                # added there own dir).
                if versionRe.search(path):
                    userDataDirDict[path] = fullPath
        # put the dirs in the correct order
        versions = userDataDirDict.keys()
        versions.sort(SmallestVersionFirst)
        userDataDirs = [userDataDirDict[version] for version in versions]
        return userDataDirs

    def _GetUserDataDirForVersion(self, major, minor):
        return os.path.join(self._userAppDataPath(),
                            "%s.%s" % (major, minor))

    def _GetCommonDataDirForVersion(self, major, minor):
        return os.path.join(self._commonAppDataPath(),
                            "%s.%s" % (major, minor))

    def get_userDataDir(self):
        infoSvc = components.classes["@activestate.com/koInfoService;1"].\
            getService(components.interfaces.koIInfoService)
        major, minor = infoSvc.version.split('.', 2)[:2]
        return self._GetUserDataDirForVersion(major, minor)

    def get_hostUserDataDir(self):
        #XXX Any change to the def'n of this directory must be carried over
        #    to komodo.c, where the logic is duplicated.
        import socket
        return os.path.join(self.get_userDataDir(),
                            "host-"+socket.gethostname())

    def get_factoryCommonDataDir(self):
        #XXX Would be better called systemCommonDataDir, or
        #    defaultSystemCommonDataDir.
        infoSvc = components.classes["@activestate.com/koInfoService;1"].\
               getService(components.interfaces.koIInfoService)
        major, minor = infoSvc.version.split('.', 2)[:2]
        return self._GetCommonDataDirForVersion(major, minor)
        
    # XXX this code is duplicated in koPrefs.py since the global prefs
    # service needs the commonDataDir, but cannot access it via
    # koIDirs due to chicken-egg issue.  If you change this, change
    # koPrefs also!
    def get_commonDataDir(self):
        globalPrefs = components.classes["@activestate.com/koPrefService;1"]\
                      .getService(components.interfaces.koIPrefService).prefs
        method = "default"
        if globalPrefs.hasStringPref("commonDataDirMethod"):
            method = globalPrefs.getStringPref("commonDataDirMethod")
        if method not in ("default", "custom"):
            log.error("bogus Common Data Dir determination method, '%s', "
                      "falling back to default", method)
        if method == "default":
            return self.get_factoryCommonDataDir()
        elif method == "custom":
            return globalPrefs.getStringPref("customCommonDataDir")

    def get_installDir(self):
        # KOMODO_INSTALLDIR is (supposed to be) set by komodo.exe
        import os
        return os.environ["KOMODO_INSTALLDIR"]

    def get_mozBinDir(self):
        # An installed Komodo can build this from the "installDir".
        # A development Komodo can just get it from the KOMODO_MOZBINDIR
        # environment variable, which, in general will not be defined for
        # installed Komodo's.
        import os
        mozBinDir = os.environ.get("KOMODO_MOZBINDIR", None)
        if mozBinDir is None:
            mozBinDir = os.path.join(self.get_installDir(), "Mozilla")
        return mozBinDir

    def get_binDBGPDir(self):
        return os.path.join(self.get_installDir(), "dbgp", "bin")
    def get_perlDBGPDir(self):
        return os.path.join(self.get_installDir(), "dbgp", "perllib")
    def get_pythonDBGPDir(self):
        return os.path.join(self.get_installDir(), "dbgp", "pythonlib")

if __name__ == "__main__":
    dirs = components.classes['@activestate.com/koDirs;1'].createInstance(components.interfaces.koIDirs)
    print dirs.userDataDir
    print dirs.commonDataDir
    try:
        print dirs.installDir
    except:
        print "couldn't get installDir -- env't variable not set right most likely"

