#!python
# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# Provide general (read-only) information about a Komodo build/installation
#
# This is accomplished by reading komodo-info.txt, distributed with
# a Komodo build/installation.

import sys
import os
import re
import strptime
import time
from xpcom import components, nsError, ServerException, COMException
import logging

log = logging.getLogger("koInfoService")

# Defines not in the nsI IDL.
PROXY_SYNC    = 0x0001  # acts just like a function call.
PROXY_ASYNC   = 0x0002  # fire and forget.  This will return immediately and
                        # you will lose all return information.
PROXY_ALWAYS  = 0x0004  # ignore check to see if the eventQ is on the same
                        # thread as the caller, and alway return a proxied
                        # object.

class KoInfoService:
    _com_interfaces_ = [components.interfaces.koIInfoService]
    _reg_clsid_ = "{EB22F329-1D99-427a-B0E1-19DFF13AFBF7}"
    _reg_contractid_ = "@activestate.com/koInfoService;1"
    _reg_desc_ = "Komodo Information Service"

    def _getInfo(self, infoURL):
        log.debug("Getting Komodo information from '%s'.\n", infoURL)
        import xpcom.file
        fin = xpcom.file.URIFile(infoURL)
        datumRe = re.compile("^(?P<name>\w+)=(?P<value>.*)$")
        info = {}
        for line in fin.readlines():
            line = line.strip()
            datumMatch = datumRe.search(line)
            if datumMatch:
                info[datumMatch.group("name")] = datumMatch.group("value")
            else:
                log.error("Bogus data line, '%s', in Komodo information "\
                          "file, '%s'.\n", line, infoURL)
                raise ServerException(nsError.NS_ERROR_UNEXPECTED)
        return info

    def __init__(self):
        self.platform = sys.platform
        self.isWindows = sys.platform.startswith("win")
        # XXX bug 33823
        # when building with gtk2, platform.py functions fail preventing
        # komodo startup.  os.uname should work fine on *nix platforms.
        if sys.platform.startswith("win"):
            import win32con
            import platform
            self.osSystem = platform.system()
            self.osRelease = platform.release()
            self.osVersion = platform.version()
            self.VER_PLATFORM_WIN32_WINDOWS = win32con.VER_PLATFORM_WIN32_WINDOWS
        else:
            self.osSystem,node,self.osRelease,self.osVersion,machine = os.uname()
        
        infoURL = "chrome://jaguar/content/resources/komodo-info.txt"
        info = self._getInfo(infoURL)
        self.version = info["version"]
        self.buildNumber = info["buildNumber"]
        self.buildASCTime = info["buildASCTime"]
        self.buildPlatform = info["buildPlatform"]
        self._releaseYear = int(info["releaseYear"])
        self._releaseMonth = int(info["releaseMonth"])
        self._releaseDay = int(info["releaseDay"])
        self.mozBinDir = info["mozBinDir"]
        self.buildType = info["buildType"]
        self.installerType = info["installerType"]
        self.productType = info["productType"]
        self.brandingTag = info["brandingTag"]
        self.brandingPhrase = info["brandingPhrase"]

        # We are in non-interactive mode if KOMODO_NONINTERACTIVE is set
        # and non-zero.
        KOMODO_NONINTERACTIVE = os.environ.get("KOMODO_NONINTERACTIVE")
        self.nonInteractiveMode = 0
        if KOMODO_NONINTERACTIVE:
            try:
                KOMODO_NONINTERACTIVE = int(KOMODO_NONINTERACTIVE)
            except ValueError:
                pass
            if KOMODO_NONINTERACTIVE:
                self.nonInteractiveMode = 1

        self.licensingType = info["licensingType"]
        self.siloedLicense = 0
        self.siloedLicenseExpires = 0
        self.siloedLicenseExpirationTime = None
        if self.licensingType == "license":
            self.siloedLicense = int(info["siloedLicense"])
            if self.siloedLicense:
                self.siloedLicenseExpires = int(info["siloedLicenseExpires"])
                if self.siloedLicenseExpires:
                    self.siloedLicenseExpirationTime = int(info["siloedLicenseExpirationTime"])

        if self.licensingType == "timebomb":
            expirationTime = int(info["timebombExpirationTime"])
            now = time.time()
            self.licenseValid = expirationTime >= now
            self.expires = 1
            self.daysUntilExpiration = int((expirationTime - now)/(60*60*24))
            self.expirationDate = time.strftime("%a %b %d, %Y",
                time.localtime(expirationTime))
            if abs(self.daysUntilExpiration) == 1:
                days = "day"
            else:
                days = "days"
            if self.daysUntilExpiration >= 0:
                self.licenseInfo = "This Komodo trial will expire on %s (%d %s left)."\
                    % (self.expirationDate, self.daysUntilExpiration, days)
            else:
                self.licenseInfo = "This Komodo trial will expire on %s (%d %s ago)."\
                    % (self.expirationDate, -self.daysUntilExpiration, days)

            self.licensee = None
            self.licenseUserName = None
            self.licenseType = None
            self.licenseToken = None
            self.licenseSerialNo = None
            self.licenseStatus = None

        elif self.licensingType == "license":
            self._featureNames = eval(info["licenseFeaturesStr_ForPython"])
            self.refreshLicenseInfo()
        else:
            log.error("Unexpected licensing type: %r", self.licensingType)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)

    def refreshLicenseInfo(self):
        """Redetermine all information relating to the installed ActiveState
        license keys.
        """
        if self.licensingType == "license":
            proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"]\
                .getService(components.interfaces.nsIProxyObjectManager)
            _licSvc = components.classes["@activestate.com/koLicense;8"]\
                .getService(components.interfaces.koILicense)
            licSvc = proxyMgr.getProxyForObject(None,
                components.interfaces.koILicense, _licSvc, PROXY_SYNC)
            
            rv, featureName, feature = licSvc.checkEx(self._featureNames,
                self._releaseYear, self._releaseMonth, self._releaseDay)
            self.licenseStatus = rv
            self.licenseValid = self.licenseStatus == licSvc.LIC_VALID
            if feature:
                self.licensee = licSvc.fetchValue(feature, "Licensee")
                self.licenseUserName = licSvc.fetchValue(feature, "UserName")
                self.licenseType = licSvc.fetchValue(feature, "LicenseType")
                self.licenseToken = licSvc.fetchValue(feature, "Token")
                self.licenseSerialNo = licSvc.fetchValue(feature, "SerialNo")
                self.daysUntilExpiration = licSvc.daysUntil(feature,
                                                            "ExpirationDate")
                if self.daysUntilExpiration == 9999:
                    self.expires = 0
                    self.expirationDate = None
                else:
                    self.expires = 1
                    expirationDate = licSvc.fetchValue(feature,
                                                       "ExpirationDate")
                    self.expirationDate = time.strftime("%a %b %d, %Y",
                        strptime.strptime(expirationDate, "%d-%m-%Y"))
            else:
                self.expires = None
                self.daysUntilExpiration = None
                self.expirationDate = None
                self.licensee = None
                self.licenseUserName = None
                self.licenseType = None
                self.licenseToken = None
                self.licenseSerialNo = None

            # Make a nice description of the license status for the user.
            if self.daysUntilExpiration and abs(self.daysUntilExpiration) == 1:
                days = "day"
            else:
                days = "days"
            if self.licenseStatus in (licSvc.LIC_VALID, licSvc.LIC_EXPIRED):
                details = ["%s license" % self.licenseType]
                if self.licenseSerialNo:
                    details.append("serial number %s" % self.licenseSerialNo)
                licensedTo = "Licensed to %s (%s)."\
                              % (self.licenseUserName or self.licensee, ", ".join(details))
            else:
                licensedTo = None

            if self.licenseStatus == licSvc.LIC_VALID:
                self.licenseInfo = licensedTo
                if self.expires:
                    self.licenseInfo += " Expires on %s (%d %s left)."\
                                        % (self.expirationDate,
                                           self.daysUntilExpiration, days)
            elif self.licenseStatus == licSvc.LIC_EXPIRED:
                self.licenseInfo = "License has expired. %s Expired on %s (%r %s ago)."\
                    % (licensedTo, self.expirationDate, -self.daysUntilExpiration, days)
            elif self.licenseStatus == licSvc.LIC_NOTCOVERED:
                self.licenseInfo = "Your license does not cover this version of Komodo."
            elif self.licenseStatus == licSvc.LIC_NOFEATURE:
                self.licenseInfo = "Your ActiveState license does not include a Komodo license."
            elif self.licenseStatus == licSvc.LIC_NOFILE:
                self.licenseInfo = "No ActiveState license file could be found."
        else:
            log.error("Unexpected licensing type: %r", self.licensingType)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)

    def get_siloedLicenseExpired(self):
        if not self.siloedLicense:
            return 0
        elif not self.siloedLicenseExpires:
            return 0
        else:
            return time.time() > self.siloedLicenseExpirationTime

    def get_siloedLicenseDaysUntilExpiration(self):
        if not self.siloedLicense and not self.siloedLicenseExpires:
            raise ServerException(nsError.NS_ERROR_FAILURE)
        else:
            secs = self.siloedLicenseExpirationTime - time.time()
            import math
            days = math.ceil(secs / 60.0 / 60.0 / 24.0)
            return days

    def windows_GetVersionEx(self):
        import win32api
        return win32api.GetVersionEx()


if __name__ == "__main__":
    info = components.classes['@activestate.com/koInfoService;1'].\
        getService(components.interfaces.koIInfoService)
    print "platform: %r" % info.platform
    if info.isWindows:
        print "VER_PLATFORM_WIN32_WINDOWS: %r" % info.VER_PLATFORM_WIN32_WINDOWS
    print "osSystem: %r" % info.osSystem
    print "osRelease: %r" % info.osRelease
    print "osVersion: %r" % info.osVersion
    print "version: %r" % info.version
    print "buildNumber: %r" % info.buildNumber
    print "buildASCTime: %r" % info.buildASCTime
    print "mozBinDir: %r" % info.mozBinDir
    print "buildType: %r" % info.buildType
    print "installerType: %r" % info.installerType
    print "productType: %r" % info.productType
    print "nonInteractiveMode: %r" % info.nonInteractiveMode
    print "licensingType: %r" % info.licensingType
    print "licenseValid: %r" % info.licenseValid
    print "licenseInfo: %r" % info.licenseInfo
    print "siloedLicense: %r" % info.siloedLicense
    print "siloedLicenseExpires: %r" % info.siloedLicenseExpires
    if info.siloedLicenseExpires:
        print "siloedLicenseExpired: %r" % info.siloedLicenseExpired
        print "siloedLicenseExpirationTime: %r" % info.siloedLicenseExpirationTime
        print "siloedLicenseDaysUntilExpiration: %r" % info.siloedLicenseDaysUntilExpiration

        print "expires: %r" % info.expires
        print "daysUntilExpiration: %r" % info.daysUntilExpiration
        print "expirationDate: %r" % info.expirationDate
    print "licensee: %r" % info.licensee
    print "licenseUserName: %r" % info.licenseUserName
    print "licenseType: %r" % info.licenseType
    print "licenseStatus: %r" % info.licenseStatus

