#!python
# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

import os
import sys
import string

import svnlib
import uriparse
import koprocessutils

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
from xpcom.server import WrapObject, UnwrapObject

import logging
log = logging.getLogger('koSVN')
log.setLevel(logging.DEBUG)

#---- support routines

def _unique(s):
    """Return a list of the elements in s, in arbitrary order, but without
    duplicates. (_Part_ of the Python Cookbook recipe.)
    """
    # get the special case of an empty s out of the way, very rapidly
    n = len(s)
    if n == 0:
        return []

    # Try using a dict first, as that's the fastest and will usually work.
    u = {}
    try:
        for x in s:
            u[x] = 1
    except TypeError:
        del u  # move on to the next method
    else:
        return u.keys()

def _splitall(path):
    """Split the given path into all its directory parts and return the list
    of those parts (see Python Cookbook recipe for test suite.
    """
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

def _commonprefix(paths):
    """An os.path.commonprefix() more suited to actual paths.
    
    It returns the common path prefix of the given paths or None, if there
    is no such common prefix.
    """
    if not paths:
        return None
    splitpaths = [_splitall(path) for path in paths]
    commonprefix = []
    for set in zip(*splitpaths):
        # This path element is part of the common prefix if it is the same
        # for every give path.
        elem = set[0]
        if sys.platform.startswith("win"):
            # Case-insensitive comparison on Windows.
            set = [p.lower() for p in set]
        if len(_unique(set)) == 1:
            commonprefix.append(elem)
        else:
            break
    if commonprefix:
        retval = os.path.join(*commonprefix)
    else:
        retval = None
    #print "_commonprefix(%r) == %r" % (paths, retval)
    return retval

# similar to listcmd.line2argv, but we keep quotes around arguments
def line2argv(line):
    line = line.strip()
    argv = []
    state = "default"
    arg = None  # the current argument being parsed
    i = -1
    while 1:
        i += 1
        if i >= len(line): break
        ch = line[i]
        
        if ch == "\\": # escaped char always added to arg, regardless of state
            if arg is None: arg = ""
            i += 1
            arg += line[i]
            continue
        
        if state == "single-quoted":
            if ch == "'":
                state = "default"
            arg += ch
        elif state == "double-quoted":
            if ch == '"':
                state = "default"
            arg += ch
        elif state == "default":
            if ch == '"':
                if arg is None: arg = ""
                state = "double-quoted"
                arg += ch
            elif ch == "'":
                if arg is None: arg = ""
                state = "single-quoted"
                arg += ch
            elif ch in string.whitespace:
                if arg is not None:
                    argv.append(arg)
                arg = None
            else:
                if arg is None: arg = ""
                arg += ch
    if arg is not None:
        argv.append(arg)
    if state != "default":
        raise ValueError("command line is not terminated: unfinished %s "
                         "segment" % state)
    return argv

def _argvFromString(options):
    options = line2argv(options)
    argv = {}
    last = None
    for o in options:
        if o[0]=='-':
            last = str(o.replace('-','_'))
            if last[0] == '_': last = last[1:]
            if last[0] == '_': last = last[1:]
            argv[last]=1
        elif last:
            argv[last] = o
    return argv

#---- component implementation

class KoSVN:
    _com_interfaces_ = [components.interfaces.koISVN, components.interfaces.nsIObserver]
    _reg_desc_ = "Komodo SVN Support"
    _reg_contractid_ = "@activestate.com/koSVN;1"
    _reg_clsid_ = "{25651630-86b7-40f5-8cad-594fd668978c}"
    
    def __init__(self):
        # generic cvs instance to do work with
        self.svn = svnlib.SVN()
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        self.fileStatusSvc = components.classes["@activestate.com/koFileStatusService;1"].\
                getService(components.interfaces.koIFileStatusService);
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self._globalPrefs.prefObserverService.addObserver(WrapObject(self,components.interfaces.nsIObserver),'svnExecutable',0)
        self._env = koprocessutils.getUserEnv()
        self._userPath = self._env["PATH"].split(os.pathsep)
        self.fileSvc = components.classes["@activestate.com/koFileService;1"].\
                getService(components.interfaces.koIFileService);

        self.reasonNotFunctional = None
        self.isFunctional = None
        self._executable = None
        self.set_executable(None)
        self._setExecutableFromPrefs()
        self.redetermineIfFunctional()
    
    def Observe(self, subject, topic, value):
        if topic == 'svnExecutable':
            self._setExecutableFromPrefs()
            
    def _setExecutableFromPrefs(self):
        executable = None
        if self._globalPrefs.hasPref('svnExecutable'):
            executable = self._globalPrefs.getStringPref('svnExecutable')
        if not executable or self._executable != executable:
            self.set_executable(executable)
            try:
                self._observerSvc.notifyObservers(self, 'subsystem_changed', 'svn')
            except COMException:
                pass # Ignore error if there is no observer.
    
    def _findExecutable(self):
        import which
        assert which._version_ >= (1,0,0)
        try:
            if sys.platform.startswith('win'):
                exts = ['.exe']
            else:
                exts = None
            svn = which.which('svn', exts=exts, path=self._userPath)
        except which.WhichError:
            svn = None
        return svn
        
    def set_executable(self, executable):
        if executable:
            self._executable = executable
        else:
            self._executable = self._findExecutable()
        if self._executable and os.path.exists(self._executable):
            self.svn.svn = self._executable
        
    def get_executable(self):
        if not self._executable:
            self._setExecutableFromPrefs()
        return self._executable
        
    def redetermineIfFunctional(self):
        if not self._executable:
            svn = self._findExecutable()
        elif os.path.exists(self._executable):
            svn = self._executable
        else:
            svn = None
        
        if svn is None:
            self.isFunctional = 0
            self.reasonNotFunctional = "no 'svn' executable could be found"
        else:
            self.isFunctional = 1
            self.reasonNotFunctional = None
        try:
            self._observerSvc.notifyObservers(self, 'subsystem_changed', 'svn')
        except COMException:
            pass # Ignore error if there is no observer.

    def _splitFile(self, file):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        try:
            path = uriparse.URIToLocalPath(file);
            if path:
                if os.path.isdir(path):
                    return path,''
                path = os.path.normpath(path)
                basedir = os.path.dirname(path)
                filename = os.path.basename(path)
                return basedir, filename
        except svnlib.SVNLibError, e:
            # we'll ignore files that are not in perforce
            self.lastErrorSvc.setLastError(0, str(e))
            pass
        return None,None
    
    def _splitFiles(self, files):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        _files = {}
        for file in files:
            basedir,filename = self._splitFile(file)
            if basedir:
                if not basedir in _files:
                    _files[basedir] = []
                _files[basedir].append(filename)
        return _files
    
    def _convertFiles(self, files):
        _files = []
        for file in files:
            path = uriparse.URIToLocalPath(file)
            path = os.path.normpath(path)
            _files.append(path)
        return _files
    
#    diff -- Display diff of client file with depot file
    def diff(self, files, options, external):
        #print "svn diff ",repr(files)
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        # convert options into a dictionary
        options = _argvFromString(options)
        if external:
            options['diff_cmd'] = external
        if 'env' not in options:
            options['env'] = self._env
        diff = ''
        for basedir, files in _files.items():
            #print "files %r basedir %r" %(files, basedir)
            for f in files:
                try:
                    result, raw = self.svn.diff(f,
                                               _raw=1,
                                               cwd=basedir,
                                               **options)
                except svnlib.SVNLibError, e:
                    self.lastErrorSvc.setLastError(0, str(e))
                    return ''
                if 'stdout' in raw:
                    diff += raw['stdout']
                elif 'stderr' in raw:
                    self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                    diff += raw['stderr']

        # Convert line-endings to the currently expected
        import eollib
        eolpref = self._globalPrefs.getStringPref('endOfLine')
        try:
            desiredEOL = eollib.eolPref2eol[eolpref]
        except KeyError:
            desiredEOL = eollib.EOL_PLATFORM
        diff = eollib.convertToEOLFormat(diff, desiredEOL)

        return diff

#    add -- Open a new file to add it to the depot
    def add(self, files, mode, message):
        #add(self, files, mode=None, msg=None, cwd=None)
        # first we figure out what repositories files are in
        # and split them up into seperate lists
        # we init the cvs object with the root we retreived
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # , filetype=mode, msg=message, cwd=basedir
            try:
                result, raw = self.svn.add(files=_files[basedir],
                                          _raw=1, cwd=basedir,
                                          env=self._env)
            except svnlib.SVNLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        # do notification that status may have changed
        # we have to use the original urls passed in since
        # they are modified to local paths in _splitFiles
        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    submit -- Submit open files to the depot
    def commit(self, files, message, options):
        #print "cvs commit ",repr(files)
        # svnlib saves message to a message for file
        self.lastErrorSvc.setLastError(0, '')

        msgfile = self.fileSvc.makeTempFile('.svn','w')
        msgfile.puts(message)
        msgfile.flush()
        msgfile.close()
        msgpath = msgfile.path
        try:
            _split = self._splitFiles(files)
            _files = self._convertFiles(files)
            basedir = _commonprefix(_split.keys())
            output = ''
            try:
                result, raw = self.svn.commit(_files,
                                             file=msgpath,
                                             _raw=1,
                                             cwd=basedir,
                                             env=self._env)
            except svnlib.SVNLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
    
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
            for file in files:
                self._observerSvc.notifyObservers(self, 'file_status_now', file)
            return output
        finally:
            del msgfile

    
#    delete -- Open an existing file to delete it from the depot
    def remove(self, files, force, recursive):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # , force, recursive,cwd=basedir
            try:
                result, raw = self.svn.delete(_files[basedir],
                                             _raw=1, cwd=basedir,
                                             env=self._env)
            except svnlib.SVNLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        for file in files:
            self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    sync -- Synchronize the client with its view of the depot
    def update(self, files, options):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # ,cwd=basedir
            try:
                result, raw = self.svn.update(_files[basedir],
                                           _raw=1,
                                           cwd=basedir,
                                           env=self._env)
            except svnlib.SVNLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
            #XXX need to deal with this
            ## build a better status output for komodo
            #if 'files' in result:
            #    for file in result['files']:
            #        localFileName = os.path.normpath(os.path.join(basedir,file['file']))
            #        url = uriparse.localPathToURI(localFileName)
            #        #print "%s %s %s" %(file['status'],file['file'],url)
            #        output += "%s %s %s\n" % (file['status'],file['file'],url)
            #output = ""
                    
        for file in files:
            self._observerSvc.notifyObservers(self, 'file_status_now', file)
        
        return output

#    revert -- Discard changes from an opened file
    def revert(self, files, options):
        # XXX unfortunately this function requires knowledge of
        # item properties
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            try:
                result, raw = self.svn.revert(_files[basedir], _raw=1,
                                             cwd=basedir,
                                             env=self._env)
            except svnlib.SVNLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        for file in files:
            self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
    def edit(self, files):
        return None

