#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys
import os
import time
import threading
import urllib
from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject

import logging
import process, koprocessutils
import uriparse
import which


log = logging.getLogger("koVPMService")
#log.setLevel(logging.DEBUG)


#---- component implementation

class KoVPMService:
    _com_interfaces_ = [components.interfaces.koIVPMService]
    _reg_clsid_ = "3B8B33E6-B813-4714-A5B0-E4D920803356"
    _reg_contractid_ = "@activestate.com/koVPMService;1"
    _reg_desc_ = "Service for interacting with the Visual Package Manager (VPM)"

    def __init__(self):
        log.debug("KoVPMService.__init__()")
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
            .getService(components.interfaces.koILastErrorService)
        proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"]\
            .getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self.observerSvc = getProxyForObject(None,
            components.interfaces.nsIObserverService,
            observerSvc, PROXY_ALWAYS | PROXY_SYNC)

        self._lock = threading.RLock()
        self._starting = 0    # True iff the server is starting up.
        self._doneStarting = threading.Event()
        self._server = None   # URL of running server or None

    def _getVPMPath(self, perlVer):
        """Return the full path the VPM executable.

        "perlVer" is the version of the Perl that will be used to launch
            VPM. On Linux this matters. Examples of the perlVer format:
            '5.005_03', '5.6.1', '5.8.0'

        If the VPM exe could not be found a suitable exception
        is raised and an error is set on koILastErrorService.
        """
        log.debug("KoVPMService._getVPMPath()")
        dirSvc = components.classes["@activestate.com/koDirs;1"]\
                 .getService(components.interfaces.koIDirs)
        dname = os.path.join(dirSvc.installDir, "vpm")
        if sys.platform.startswith("win"):
            fname = os.path.join(dname, "vpm.exe")
        else:
            if perlVer.startswith("5.6"):
                fname = os.path.join(dname, "vpm-5.6")
            elif perlVer.startswith("5.8"):
                fname = os.path.join(dname, "vpm-5.8")
            else:
                fname = os.path.join(dname, "vpm-"+perlVer[:3])
        if os.path.exists(fname):
            return fname
        else:
            ex = "Could not find the VPM executable: '%s' does not exist" % fname
            self.lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

    def _getAppropriatePerlInfo(self):
        """Return the 2-tuple (<full-path to Perl bin dir>, <Perl version>)
        for the appropriate Perl installation.
        
        "Appropriate" means:
            - Choose the one specifically selected by the user for Perl
              debugging if one is so chosen.
            - Otherwise use 'which' to find available Perl installations.
            - Ensure that perl is at least ActivePerl 633.
        XXX This is a weak way of determining the Perl for VPM to act on.
            The user could, I imagine, be surprised by which Perl is selected
            and frustrated that that perl selection cannot easily be
            controlled.

        Raises an exception and sets the last error with an explanation if
        no appropriate Perl could be found.
        """
        log.debug("KoVPMService._getAppropriatePerlInfo()")
        minBuildNumber = 633
        perlInfo = components.classes["@activestate.com/koPerlInfoEx;1"]\
                   .createInstance(components.interfaces.koIAppInfoEx)

        # Is the user's selected Perl sufficient, if selected.
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                  .getService(components.interfaces.koIPrefService)
        perlDefaultInterp = prefSvc.prefs.getStringPref("perlDefaultInterpreter")
        if perlDefaultInterp:
            installdir = os.path.dirname(os.path.dirname(perlDefaultInterp))
            perlInfo.installationPath = installdir
            exePath = perlInfo.executablePath
            if not os.path.exists(exePath):
                msg = "Perl installation does not exist: \"%s\"" % exePath
                self.lastErrorSvc.setLastError(0, msg)
                raise ServerException(nsError.NS_ERROR_UNEXPECTED, msg)

            try:
                buildNumber = perlInfo.buildNumber
            except COMException, ex:
                msg = "Perl installation is not ActivePerl: %s" % installdir
                self.lastErrorSvc.setLastError(0, msg)
                raise ServerException(nsError.NS_ERROR_UNEXPECTED, msg)

            if buildNumber >= minBuildNumber:
                return (os.path.join(installdir, "bin"),
                        perlInfo.version)

        # Look on PATH.
        userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)
        if sys.platform.startswith('win'):
            exts = ['.exe']
        else:
            exts = None
        perls = which.whichall('perl', exts=exts, path=userPath)
        for perl in perls:
            installdir = os.path.dirname(os.path.dirname(perl))
            perlInfo.installationPath = installdir
            try:
                buildNumber = perlInfo.buildNumber
            except COMException, ex:
                # This is probably a non-ActivePerl build.
                continue
            if buildNumber >= minBuildNumber:
                return (os.path.join(installdir, "bin"),
                        perlInfo.version)
        else:
            ex = "Could not find a suitable Perl installation on which to "\
                 "run VPM. VPM currently requires ActivePerl build %d "\
                 "or later." % minBuildNumber
            self.lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

    def get_isFunctional(self):
        log.debug("KoVPMService.get_isFunctional()")
        try:
            perlDir, perlVer = self._getAppropriatePerlInfo()
            self._getVPMPath(perlVer)
        except Exception:
            return 0
        else:
            return 1

    def _notify(self, url):
        log.debug("KoVPMService._notify()")
        try:
            self.observerSvc.notifyObservers(None, "vpm_server_started", url)
        except COMException, e:
            # do nothing: Notify sometimes raises an exception if
            # receivers are not registered
            pass

    def _startAndNotify(self, perlDir, vpm):
        log.debug("KoVPMService._startAndNotify()")
        env = koprocessutils.getUserEnv()
        # VPM will operate on the first Perl in the user's path so put the
        # "appropriate" perl first on the PATH.
        if env.has_key("PATH"):
            env["PATH"] = perlDir + os.pathsep + env["PATH"]
        else:
            env["PATH"] = perlDir
        argv = [vpm, '--showurl', '--noquit']

        p = process.ProcessOpen(argv, env=env)
        output = p.stdout.read()
        if not output.startswith("http://"):
            error = p.stderr.read()
        else:
            # VPM hangs if you try to read stderr if there is none.
            error = None
        p.close()
        
        self._lock.acquire()
        url = output.strip()
        if error:
            # There likely was an error starting VPM. Get the error message
            # and pass it back so it can be reported to the user.
            url = error
        else:
            self._server = url
        self._starting = 0
        self._doneStarting.set()
        self._lock.release()

        self._notify(url)

    def _notifyWhenStarted(self):
        log.debug("KoVPMService._notifyWhenStarted()")
        self._doneStarting.wait()
        url = self._getServerURL()
        if url:
            self._notify(url)

    def _getServerURL(self):
        """Return the URL of an already running VPM server if there is one."""
        log.debug("KoVPMService._getServerURL()")
        self._lock.acquire()
        retval = None
        if self._server:
            try:
                # If this fails the server is no longer operational.
                urllib.urlretrieve(self._server)
            except IOError:
                self._server = None
            else:
                retval = self._server
        self._lock.release()
        return retval

    def start(self):
        log.info("KoVPMService.start()")
        msg = None  # Message for the user.

        # Determine what to do.
        self._lock.acquire()
        action = None
        url = self._getServerURL()
        if url:
            action = "use-existing-server"
        elif self._starting:
            action = "use-existing-server-when-started"
        else:
            self._starting = 1
            self._doneStarting.clear()
            action = "start-new-server"
        self._lock.release()

        # Do it.
        if action == "use-existing-server":
            self._notify(url)
        elif action == "use-existing-server-when-started":
            t = threading.Thread(target=self._notifyWhenStarted)
            t.start()
            msg = "The VPM server is already starting. This may take a "\
                  "few minutes."
        elif action == "start-new-server":
            try:
                perlDir, perlVer = self._getAppropriatePerlInfo()
                vpm = self._getVPMPath(perlVer)
            except ServerException:
                self._lock.acquire()
                self._starting = 0
                self._doneStarting.set()
                self._lock.release()
                raise
            t = threading.Thread(target=self._startAndNotify,
                                 kwargs={'perlDir': perlDir, 'vpm': vpm})
            t.start()
        else:
            raise ServerException(nsError.NS_ERROR_UNEXPECTED,
                                  "Unexpected action: '%s'" % action)
        return msg

    def shutdown(self):
        log.info("KoVPMService.shutdown()")
        if self._server:
            #print "Shutdown VPM server at: %s" % url
            try:
                urllib.urlretrieve(self._server+"/quit")
            except IOError:
                # Ignore errors possible if the VPM was already shutdown
                # otherwise. E.g.
                #   IOError: [Errno socket error] (10061, 'Connection refused')
                pass


