#!python
# Copyright (c) 2004 ActiveState Corp.
# See the file LICENSE.txt for licensing information.

"""Code Intel Type Determination Lang (CITDL, "citadel") processing."""

import os
import sys
import time
import md5
import pprint
from sets import Set
import bisect
import glob
import weakref

import logging
import process

from codeintel.common import *  # ScanRequest and exceptions are defined here


#---- global variables

log = logging.getLogger("citdl")
#log.setLevel(logging.INFO)
CACHING = 1

# CIDB table row indeces.
(S_ID, S_FILE_ID, S_SCAN_ID, S_MODULE_ID, S_SCOPE_ID, S_SCOPE, S_NAME,
 S_SIGNATURE, S_LINE, S_LINEEND, S_TYPE, S_ATTRIBUTES, S_DOC) = range(13)
(M_ID, M_FILE_ID, M_SCAN_ID, M_NAME, M_LINE, M_ATTRIBUTES, M_DOC) = range(7)
(I_ID, I_FILE_ID, I_SCAN_ID, I_MODULE_ID, I_SCOPE_ID, I_LINE, I_MODULE,
 I_SYMBOL, I_ALIAS) = range(9)


#---- Import handlers (import semantics vary among languages)

def _preferredCallTipsFirst(a, b):
    # - prefer calltips with a newline (includes a desc string)
    aHasEOL = '\n' in a
    bHasEOL = '\n' in b
    if aHasEOL and not bHasEOL:
        return -1
    elif not aHasEOL and bHasEOL:
        return 1
    else:
        #XXX:TODO if necessary:
        # - prefer calltips that start with the expr symbol ref
        return 0

def getImportHandler(language):
    lang2handler = {
        "Python": PythonImportHandler,
        "Perl": PerlImportHandler,
        "PHP": PHPImportHandler,
        "JavaScript": JavaScriptImportHandler,
        "Tcl": TclImportHandler,
    }
    if language not in lang2handler:
        raise CodeIntelError("cannot handle '%s' imports: there is no "
                             "registered ImportHandler class for "
                             "language '%s'" % (language, language))
    return lang2handler[language]()

class ImportHandler:
    """Virtual base class for language-specific "import"-statement handlers.
    
    The basic job of an import handler is to convert an import statement (i.e.
    a row in the 'import' table) into a row in the CIDB 'module' table. Doing
    this depends on language-specific import semantics.
    
    A fundamental part of import resolution is the search path. Here the
    search path is broken into three parts:
        - "core" path: built-in to the interpreter/compiler, generally
          dependent on the installation location.
        - "env" path: additional directories specified in a special
          environment variable, e.g. PYTHONPATH, PERL5LIB
        - "custom" path: additional "out-of-band" directories
    
    Each language-specific ImportHandler is a singleton, as controlled by
    Manager.getImportHandler().
    """
    PATH_ENV_VAR = None
    corePath = None
    envPath = None
    customPath = None

    def setCustomPath(self, path):
        """Specify some custom search directories."""
        self.customPath = path
    def setEnvPath(self, value=None):
        """Specify the value of the PATH-style environment variable.
        
            "value" is the appropriate environment variable value, e.g.:
                "C:\trentm\mylib;C:\shared\lib-python". If value is None then
                the value will be retrieved from os.environ.

        This will lazily be called if necessary.
        """
        path = []
        if value is None:
            if self.PATH_ENV_VAR:
                path = os.environ.get(self.PATH_ENV_VAR, "").split(os.pathsep)
        else:
            path = value.split(os.pathsep)
        self.envPath = path
    def setCorePath(self, compiler=None, extra=None):
        """Specify the data needed to determine the core search path.
        
            "compiler" is the path to the language compiler/interpreter.
                If not specified then the first appropriate
                compiler/interpreter on the path is used.
            "extra" (optional) can be used to specify required extra
                data for determining the core import path.  For example,
                the PHP-specific implementation uses this to specify the
                "php.ini"-config-file path.

        This will lazily be called if necessary.
        """
        # Sub-classes must implement this and set self.corePath as a
        # result (to [] if it could not be determined).
        raise NotImplementedError("setCorePath: pure virtual method call")

    def _getPath(self, cwd=None):
        """Put all the path pieces together and return that list.
        
        If "cwd" is specified, it is prepended to the list. (In many languages
        the directory of the file with the import statement is first on the
        module search path.)
        """
        if self.corePath is None: self.setCorePath()
        if self.envPath is None: self.setEnvPath()
        if cwd is not None:
            path = [cwd]
        else:
            path = []
        if self.customPath:
            path += self.customPath
        path += self.envPath
        path += self.corePath
        return path

    def findModuleOnDisk(self, module, submodule, cwd):
        """Find the full path to the given module on disk.
        
            "module" is a the module name (_can_ include dots, e.g. xml.sax)
            "submodule" is a possible sub-module to also include in the
                import path. Some languages allow you to separate the module
                and submodule tokens to control the imported symbol name. For
                example, submodule will be "sax" for this Python import:
                    from xml import sax
                This kind of thing is not possible in Perl and hence the
                "submodule" argument can be ignored for a Perl
                implementation. Submodule would be "DumperX" from this Perl
                import:
                    use Data::Dumper qw(DumperX);
            "cwd" is the directory of the module containing the import stmt.
        
        If successful, returns a 3-tuple:
            (<module path>, "module"|"submodule", <scannable>,
        where,
            the second element indicates if the returned module row
                corresponds to the given "module" or "submodule" argument.
            <scannable> is a boolean indicating if this file is likely
                scannable.

        Otherwise, returns (None, None, None).
        """
        raise NotImplementedError("findModuleOnDisk: pure virtual method call")

    def findSubImportsOnDisk(self, module, cwd):
        """Return a list of importable submodules to the given module.
        
        The returned list is a list of strings that would be appropriate for
        direct use in the language's specific import statement. I.e.: for
        Perl:
            ["ConnCache", "Protocol", "UserAgent", ...]
        not:
            ["ConnCache.pm", "Protocol.pm", "UserAgent.pm", ...]
        For PHP, however, including the .php extension might be appropriate.
        """
        raise NotImplementedError("findSubImportsOnDisk: pure virtual method call")

    def findModule(self, cu, factory, module, submodule, cwd=None):
        """Find the 'module' table row corresponding to the given import row.
        
            "cu" is an open CIDB cursor
            "factory" is a ThingyFactory instance to use.
            "module" is the module name (or path, e.g. "xml.sax")
            "submodule" is a possible sub-module to also include in the
                import path. Some languages allow you to separate the module
                and submodule tokens to control the imported symbol name. For
                example, submodule will be "sax" for this Python import:
                    from xml import sax
                This kind of thing is not possible in Perl and hence the
                "submodule" argument can be ignored for a Perl
                implementation. Submodule would be "DumperX" from this Perl
                import:
                    use Data::Dumper qw(DumperX);
            "cwd" is the directory of the module containing the import stmt.

        If successful, returns a 2-tuple:
            (<module Scope instance>, "module"|"submodule")
        The second element indicates if the returned module row corresponds
        to the given "module" or "submodule" argument.
        
        XXX To re-scan or not to re-scan:
            May want to schedule a re-scan of the file whether it is in the
            CIDB or not. Or maybe just if it isn't. Or maybe this is up to
            the CIX importing to kick off and we don't need to worry about it.
            In any case, if we _do_ re-scan files already in the CIDB then
            we might want to exclude files whose "md5" value is marked with
            the special "*", meaning: "don't re-scan".
        """
        raise NotImplementedError("findModule: pure virtual method call")

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        """Generate scannable files on the import path.
        
            "path" (optional) is an import path to load. If not specified
                the default import path is used.
            "skipRareImports" (optional, default false) is a boolean
                indicating if files unlikely to be imported/searched-for/used
                should be skipped. This can be specified to speed up, for
                example, scanning *all* files during a batch update of a
                language installation.
            "importableOnly" (optional, default false) is a boolean
                indicating if only those files that are importable from
                the given path should be included. For example a Python
                file in a subdirectory cannot be imported if that dir
                does not have a package-defining "__init__.py" file.
        """
        raise NotImplementedError("genScannableFiles: pure virtual method call")



class PythonImportHandler(ImportHandler):
    PATH_ENV_VAR = "PYTHONPATH"

    def __init__(self):
        self.__stdCIXScanId = None

    def _shellOutForPath(self, compiler):
        import process
        argv = [compiler, "-c", "import sys; print '\\n'.join(sys.path)"]
        # Can't use -E to ignore PYTHONPATH because older versions of
        # Python don't have it (e.g. v1.5.2).
        env = dict(os.environ)
        if "PYTHONPATH" in env: del env["PYTHONPATH"]
        if "PYTHONHOME" in env: del env["PYTHONHOME"]
        if "PYTHONSTARTUP" in env: del env["PYTHONSTARTUP"]

        p = process.ProcessOpen(argv, env=env)
        retval = p.wait()
        output = p.stdout.read()
        p.close()
        path = [line for line in output.splitlines(0)]
        if path and (path[0] == "" or path[0] == os.getcwd()):
            del path[0] # cwd handled separately
        return path

    def setCorePath(self, compiler=None, extra=None):
        if compiler is None:
            import which
            assert which._version_ >= (1, 0, 0)
            compiler = which.which("python")
        self.corePath = self._shellOutForPath(compiler)

    def _getStdCIXScanId(self, cu):
        if not self.__stdCIXScanId:
            cu.execute("SELECT scan.id FROM scan, file, language "
                       " WHERE language.name='Python'"
                       "   AND language.id=file.language_id"
                       "   AND scan.file_id=file.id"
                       "   AND scan.generator='StdCIX' LIMIT 1")
            for row in cu:
                self.__stdCIXScanId = row[0]
                break
        return self.__stdCIXScanId

    def findModule(self, cu, factory, moduleName, submodule, cwd=None):
        """Algorithm:
        - Look for the appropriate file on disk, using the import path.
        - If found, see if we have this file in the CIDB.
            - If it's in the CIDB, then find the module with the appropriate
              name from that file.
        - If haven't found a matching module row yet, then lookup the full
          import path (and language) in the 'module' table and get all hits.
          This has the potential to "hit" modules that aren't in the search
          path, but a false positive is better than nothing in this case.
        - If still no hits, then raise an error that describes the problem
          in as helpful a way as possible for the user.
        """
        module = kind = None

        if CACHING:
            key = (moduleName, submodule, cwd)
            retval = factory.pythonFindModuleCache.get(key, None)
            if retval is not None:
                return retval

        # Some specific modules are much better defined in 'python.cix'
        # than a scan of the associated .py file.
        #XXX The _correct_ algorithm would probably be to (1) find the
        #    proper file to import via findModuleOnDisk and (2) if that
        #    file was from the stdlib that the overrides represent, only
        #    _then_ would the override take effect. Currently this will
        #   override a user's own os.py. This is not deemed a big deal.
        overrides = ["os", "os.path", "re"]
        if moduleName in overrides:
            stdCIXScanId = self._getStdCIXScanId(cu)
            if stdCIXScanId:
                sql = "SELECT * FROM module WHERE scan_id=? AND name=? LIMIT 1"
                if submodule:
                    cu.execute(sql, (stdCIXScanId, moduleName+'.'+submodule))
                    for row in cu:
                        module = factory.createScope(row[M_FILE_ID], "module",
                                                     row[M_ID], row)
                        kind = "submodule"
                        break
                if not module:
                    cu.execute(sql, (stdCIXScanId, moduleName))
                    for row in cu:
                        module = factory.createScope(row[M_FILE_ID], "module",
                                                     row[M_ID], row)
                        kind = "module"
                        break
        
        if not module:
            # Look for the appropriate file on disk, using the import path.
            modfile, kind, scannable = self.findModuleOnDisk(moduleName,
                                                             submodule, cwd)

            # If found, see if we have this file and module in the CIDB.
            file_id = None
            if modfile:
                if kind == "module":
                    modname = moduleName.split('.')[-1]
                elif kind == "submodule":
                    modname = submodule
                cu.execute("SELECT * FROM module "
                           "WHERE file_id=(SELECT id FROM file WHERE compare_path=?) "
                           "AND name=? LIMIT 1",
                           (canonicalizePath(modfile), modname))
                for row in cu:
                    module = factory.createScope(row[M_FILE_ID], "module",
                                                 row[M_ID], row)
                    break

        if CACHING and module:
            # Only insert into the cache here, and NOT if the retval
            # is gotten via the fallback.
            factory.pythonFindModuleCache[key] = (module, kind)
            factory.pythonFindModuleCacheIndex[module.file_id] = key

        # Fallback: If haven't found a matching module row yet, then lookup
        # the full import path (and language) in the 'module' table and get
        # all hits. First one wins.
        if not module:
            #XXX Might want database index specifically for this?
            sql = "SELECT module.* FROM language, module, file "\
                  " WHERE language.name='Python' "\
                  "   AND language.id=file.language_id "\
                  "   AND module.file_id=file.id "\
                  "   AND module.name=? LIMIT 1"
            if submodule:
                cu.execute(sql, (moduleName+'.'+submodule,))
                for row in cu:
                    module = factory.createScope(row[M_FILE_ID], "module",
                                                 row[M_ID], row)
                    kind = "submodule"
                    break
            if not module:
                cu.execute(sql, (moduleName,))
                for row in cu:
                    module = factory.createScope(row[M_FILE_ID], "module",
                                                 row[M_ID], row)
                    kind = "module"
                    break

        if not module:
            if not submodule:
                desc = "'%s' module" % moduleName
            else:
                desc = "'%s.%s' or '%s' module" % (moduleName, submodule, moduleName)
            raise CodeIntelError("couldn't find %s in database" % desc)
        else:
            return (module, kind)

    def findModuleOnDisk(self, module, submodule, cwd=None):
        r"""        
            >>> h = PythonImportHandler()
            >>> h.findModuleOnDisk("cmd", None)
            ('C:\\Python22\\lib\\cmd.py', 'module', 1)
            >>> h.findModuleOnDisk("xml", None)
            ('C:\\Python22\\lib\\xml\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml", "sax")
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'submodule', 1)
            >>> h.findModuleOnDisk("xml", "*")
            ('C:\\Python22\\lib\\xml\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml", "__version__")
            ('C:\\Python22\\lib\\xml\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml.sax", None)
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml.sax", "handler")
            ('C:\\Python22\\lib\\xml\\sax\\handler.py', 'submodule', 1)
            >>> h.findModuleOnDisk("xml.sax", "parse")
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml.sax", "*")
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("_tkinter", "*")
            ('C:\\Python22\\DLLs\\_tkinter.pyd', 'module', 0)
        """
        if isinstance(cwd, unicode):
            try:
                cwd = str(cwd) # imp.find_module() doesn't like unicode paths
            except:
                log.warn("could not convert unicode CWD to string for Python"
                         "module lookup: lookup might fail: cwd=%r" % cwd)
                cwd = None
        path = self._getPath(cwd)

        import imp
        try:
            modparts = module.split('.')
            modpath = description = None
            for i in range(len(modparts)):
                modpart = modparts[i]
                fin, modpath, desc = imp.find_module(modpart, path)
                if fin: fin.close()
                path = [modpath]
            kind = "module"
            
            # Check to see if the given sub-module can be tacked on.
            if submodule:
                try:
                    sfin, smodpath, sdesc = imp.find_module(submodule, [modpath])
                    if sfin: sfin.close()
                except ImportError:
                    pass
                else:
                    kind = "submodule"
                    modpath = smodpath
                    desc = sdesc
    
            # If the "module path" is a package directory, then convert it to
            # the appropriate _file_.
            if desc[2] == imp.PKG_DIRECTORY:
                fin, modpath, desc = imp.find_module("__init__", [modpath])
    
            # The import is scannable if it is a plain Python source file.
            suffix, mode, type = desc
            scannable = type == imp.PY_SOURCE
    
            return (modpath, kind, scannable)
        except ImportError:
            return (None, None, None)

    def _findScannableFiles(self,
                            (files, searchedDirs, skipRareImports,
                             importableOnly),
                            dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        if skipRareImports:
            if (os.path.basename(dirname) == "encodings"
                and "undefined.py" in names):
                # Skip most of the specific encoding definitions (saves
                # about 50 files).
                names = [n for n in names if n == "__init__.py"
                         or os.path.splitext(n)[0].endswith("_codec")]
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                if skipRareImports:
                    # Skip Python's test package (saves over 200 files)
                    # and other likely test dirs.
                    if names[i] in ("test", "tests"):
                        del names[i]
                        continue
                if importableOnly:
                    possibles = [os.path.join(path, "__init__.py"),
                                 os.path.join(path, "__init__.pyc"),
                                 os.path.join(path, "__init__.pyo")]
                    for possible in possibles:
                        if os.path.isfile(possible):
                            break
                    else:
                        del names[i] # don't traverse non-package dirs
                        continue
                if path.endswith(os.path.join("win32com", "gen_py")):
                    del names[i]
                    continue
            elif os.path.splitext(names[i])[1] in (".py", ".pyw"):
                #XXX The list of Python extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for "Python" files should probably include
                #    python scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs, skipRareImports,
                          importableOnly))
            for file in files:
                yield file



class PerlImportHandler(ImportHandler):
    PATH_ENV_VAR = "PERL5LIB"

    # Try to speed up self._getPath() a little bit. This is called
    # *very* frequently for Perl.
    def __init__(self):
        self._pathCache = None
        self._findModuleOnDiskCache = {}
    if CACHING:
        def _getPath(self, cwd=None):
            if self._pathCache is None:
                self._pathCache = ImportHandler._getPath(self) # intentionally exclude cwd
            if cwd:
                return [cwd] + self._pathCache
            else:
                return self._pathCache

    def _shellOutForPath(self, compiler):
        import process
        sep = "--WomBa-woMbA--"
        argv = [compiler, "-e", "print join('%s', @INC);" % sep]
        env = dict(os.environ)
        if "PERL5LIB" in env: del env["PERL5LIB"]
        if "PERLLIB" in env: del env["PERLLIB"]

        p = process.ProcessOpen(argv, env=env)
        retval = p.wait()
        output = p.stdout.read()
        error = p.stderr.read()
        retval = p.wait()
        p.close()
        if retval:
            raise CodeIntelError("could not determine Perl import path: %s"
                                 % error)
        path = [os.path.normpath(d) for d in output.split(sep)]
        # cwd handled separately
        path = [p for p in path if p not in (os.curdir, os.getcwd())]
        return path

    def setCorePath(self, compiler=None, extra=None):
        if compiler is None:
            import which
            assert which._version_ >= (1, 0, 0)
            compiler = which.which("perl")
        self.corePath = self._shellOutForPath(compiler)

    def findModule(self, cu, factory, moduleName, dummy, cwd=None):
        log.debug("   PerlImportHandler.findModule(module=%r, "
                  "submodule=%r, cwd=%r)", moduleName, dummy, cwd)
        
        if CACHING:
            key = (moduleName, cwd)
            retval = factory.perlFindModuleCache.get(key, None)
            if retval is not None:
                return retval

        # Look for the appropriate file on disk, using the import path.
        modfile, kind, scannable = self.findModuleOnDisk(moduleName, dummy, cwd)
        
        # If found, see if we have this file and module in the CIDB.
        module = file_id = None
        if modfile:
            cu.execute("SELECT * FROM module "
                       "WHERE file_id=(SELECT id FROM file WHERE compare_path=?) "
                       "AND name=? LIMIT 1",
                       (canonicalizePath(modfile), moduleName))
            for row in cu:
                module = factory.createScope(row[M_FILE_ID], "module",
                                             row[M_ID], row)
                break

        if CACHING and module:
            # Only insert into the cache here, and NOT if the retval
            # is gotten via the fallback because that probably means the
            # module just isn't scanned yet.
            # - There are exceptions, e.g. HTML::FormatText, HTML::FormatPS,
            #   that defy existence.
            factory.perlFindModuleCache[key] = (module, "module")
            factory.perlFindModuleCacheIndex[module.file_id] = key

        ## If haven't found a matching module row yet then just look for any
        ## loaded Perl module of the same name. First one wins.
        ## PERF: Is this crucial? I don't think it is. Because of that and
        ##       because it is difficult to cache the case of modules that
        ##       are NEVER found (e.g. HTML::FormatText) this is currently
        ##       being removed.
        #if not module:
        #    sql = "SELECT module.* FROM language, file, module"\
        #          " WHERE language.name='Perl' "\
        #          "   AND language.id=file.language_id "\
        #          "   AND module.file_id=file.id "\
        #          "   AND module.name=? LIMIT 1"
        #    cu.execute(sql, (moduleName,))
        #    for row in cu:
        #        module = factory.createScope(row[M_FILE_ID], "module",
        #                                     row[M_ID], row)
        #        kind = "module"
        #        break

        log.debug("   PerlImportHandler.findModule: '%s' -> %r",
                  moduleName, module)
        if not module:
            raise CodeIntelError("couldn't find '%s' module in database"
                                 % moduleName)
        else:
            return (module, "module")

    def findModuleOnDisk(self, module, dummy, cwd=None):
        #XXX Should handle some of the weird "module" values possible with
        #    Perl require statements:
        #       require 'MyModule.pm';      # module: 'MyModule.pm'
        #       require "MyModule.pm";      # module: "MyModule.pm"
        #       require $filename;          # module: $filename
        
        if CACHING:
            # The findModuleOnDisk cache is a time-based cache. I.e. it make
            # the presumption that the file system hasn't changed
            # significantly in the last N seconds.
            N = 300 # 5 minutes
            key = (module, cwd)
            if key in self._findModuleOnDiskCache\
               and self._findModuleOnDiskCache[key][0] > time.time() - N:
                return self._findModuleOnDiskCache[key][1]

        path = self._getPath(cwd)
        mname = module.replace("::", os.sep) + ".pm"
        for dname in path:
            mpath = os.path.join(dname, mname)
            if os.path.exists(mpath):
                break
        else:
            log.debug("   PerlImportHandler.findModuleOnDisk: could not "
                      "find '%s' in path: %s", module, path)
            retval = (None, None, None)
            if CACHING:
                self._findModuleOnDiskCache[key] = (time.time(), retval)
            return retval
        log.debug("   PerlImportHandler.findModuleOnDisk: '%s' -> '%s'",
                  module, mpath)
        retval = (mpath,
                  "module", # The "submodule"-thing isn't relevant in Perl
                  True)     # only handling .pm files, so always scannable
        if CACHING:
            self._findModuleOnDiskCache[key] = (time.time(), retval)
        return retval

    def findSubImportsOnDisk(self, module, cwd):
        path = self._getPath(cwd)
        mrelpath = module.replace("::", os.sep)
        subimports = {} # use a dict to get a unique list
        for dname in path:
            submodules = glob.glob(os.path.join(dname, mrelpath, "*.pm"))
            for submodule in submodules:
                subimport = os.path.splitext(os.path.basename(submodule))[0]
                subimports[subimport] = 1
        return subimports.keys()

    def _findScannableFiles(self,
                            (files, searchedDirs, skipTheseDirs, skipRareImports),
                            dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        if skipRareImports:
            # Skip .pl files when scanning a Perl lib/sitelib.
            scannableExts = (".pm",)
        else:
            scannableExts = (".pl", ".pm")
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                if os.path.normcase(path) in skipTheseDirs:
                    del names[i]
                elif skipRareImports and not ('A' <= names[i][0] <= 'Z'):
                    # Perl good practice dictates that all module directories
                    # begin with a capital letter. Therefore, we skip dirs
                    # that start with a lower case.
                    del names[i]
            elif os.path.splitext(names[i])[1] in scannableExts:
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                # Do NOT traverse the common '.' element of @INC. It is
                # environment-dependent so not useful for the typical call
                # of this method.
                continue
            skipTheseDirs = [os.path.join(dirname, "auto")]
            skipTheseDirs = [os.path.normcase(d) for d in skipTheseDirs]
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs, skipTheseDirs,
                          skipRareImports))
            for file in files:
                yield file


class PHPImportHandler(ImportHandler):
    def setCorePath(self, compiler=None, extra=None):
        #XXX To do this independent of Komodo this would need to do all
        #    the garbage that koIPHPInfoEx is doing to determine this. It
        #    might also require adding a "rcfile" argument to this method
        #    so the proper php.ini file is used in the "_shellOutForPath".
        #    This is not crucial now because koCodeIntel._Manager() handles
        #    this for us.
        if not self.corePath:
            raise CodeIntelError("Do not know how to determine the core "
                                 "PHP include path. 'corePath' must be set "
                                 "manually.")

    def _findScannableFiles(self, (files, searchedDirs), dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                pass
            elif os.path.splitext(names[i])[1] in (".php", ".inc"):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                # Do NOT traverse '.' if it is in the include_path. Not sure
                # if this is at all common for PHP.
                continue
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs))
            for file in files:
                yield file


class JavaScriptImportHandler(ImportHandler):
    def setCorePath(self, compiler=None, extra=None):
        self.corePath = []

    def _findScannableFiles(self, (files, searchedDirs), dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                pass
            elif os.path.splitext(names[i])[1] in (".js",):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                # Do NOT traverse the common '.' element of @INC. It is
                # environment-dependent so not useful for the typical call
                # of this method.
                continue
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs))
            for file in files:
                yield file


class TclImportHandler(ImportHandler):
    # Tcl _does_ have a TCLLIBPATH environment variable for specifying
    # import path elements, but parsing isn't straighforward -- it uses
    # Tcl-syntax -- so we don't bother to separate "envPath" and "corePath"
    # for Tcl.
    PATH_ENV_VAR = None

    def _shellOutForPath(self, compiler):
        import process
        argv = [compiler]
        p = process.ProcessOpen(argv)
        p.stdin.write("puts [join $auto_path \\n]")
        p.stdin.close()
        retval = p.wait()
        output = p.stdout.read()
        p.close()
        path = [os.path.normpath(line) for line in output.splitlines(0)]
        if path and (path[0] == "" or path[0] == os.getcwd()):
            del path[0] # cwd handled separately
        return path

    def setCorePath(self, compiler=None, extra=None):
        if compiler is None:
            import which
            assert which._version_ >= (1, 0, 0)
            compiler = which.which("tclsh")
        self.corePath = self._shellOutForPath(compiler)

    def _findScannableFiles(self, (files, searchedDirs, skipRareImports),
                            dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                pass
            elif os.path.splitext(names[i])[1] in (".tcl",):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                if skipRareImports and names[i] == "pkgIndex.tcl":
                    continue
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                continue
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs, skipRareImports))
            for file in files:
                yield file



#---- CITDL processing stuff

class Thingy:
    """Represents a module or symbol object/scope or import in the CIDB."""
    table = None
    id = None
    row = None

    def __init__(self, factory, file_id, table, id, row=None, cu=None):
        self.factory = factory
        # The 'row' can be passed in or a CIDB cursor, 'cu', can be passed
        # in and it will be retrieved.
        self.file_id = file_id
        self.table = table
        self.id = id
        if row is None:
            cu.execute("SELECT * FROM %s WHERE id=%d LIMIT 1"
                       % (self.table, self.id))
            self.row = cu.next()
        else:
            self.row = row
        self.cache = {}

    def guid(self):
        return (self.table, self.id)

    def __repr__(self):
        details = ""
        if self.row:
            if self.table == "module":
                details = " %s" % self.row[M_NAME]
            elif self.table == "symbol":
                details = " %s %s" % (symbolType2Name(self.row[S_TYPE]),
                                      self.row[S_NAME])
            elif self.table == "import":
                details = ' '+str(self)
        klass = str(self.__class__).split('.')[-1]
        s = "<%s %s %s:%s>" % (klass, self.table, self.id, details)
        return s

    def __nonzero__(self):
        return self.table is not None and self.id is not None

    def getBaseScope(self, cu, language):
        """Return an appropriate scope in which to resolve base class (or
        base interface) references.
        
            "cu" is an open CIDB cursor.
            "language" is the current scope file language.

        Perl seems to be special is that the scope in which base classes are
        defined, e.g.:
            package Foo;
            @ISA = qw(Bar);
        is the scope of the class/package itself, rather than the parent
        scope, e.g.:
            class Foo(Bar): ...             # Python
            class Foo extends Bar { ...     # PHP
        (Not sure about other languages.)
        """
        if language == "Perl":
            # Basically we are returning 'self', but we transform it to
            # a Scope instance because there is a slight difference between
            # "package Foo the class object" and "package Foo the scope"
            # in _resolveGetattr() evaluation.
            return self.factory.createScope(self.file_id, self.table,
                                            self.id, self.row)
        else:
            return self.getParentScope(cu, language)

    def getParentScope(self, cu, language):
        """Return a parent scope object for this scope.
        
            "cu" is an open CIDB cursor.
            "language" is the current scope file language.
        
        "Scoping" objects are modules, classes (namespaces are just a special
        kind of class) and functions. The CodeIntel system doesn't support
        lexical scopes created for if-, while-, for-, and plain brace-blocks.
        
        The parent of a class method is the class scope. The parent of a
        top-level class or function is its module. The parent of a module
        is the special built-in module (named "*"), if there is one.
        
        Returns None if there is no parent scope.
        """
        if self.table == "import":
            raise CodeIntelError("shouldn't be calling getParentScope on an"
                                 "Import instance")
        elif self.table == "module" and self.row[M_NAME] == "*":
            scope = None  # no parent of the built-in module
        elif self.table == "module":
            scope = self.factory.createBuiltInModuleScope(cu, language)
        elif not self.row[S_SCOPE_ID]:
            module_id = int(self.row[S_MODULE_ID])
            rows = self.query(cu,
                "SELECT * FROM module WHERE id=%d LIMIT 1" % module_id)
            if not rows:
                raise CIDBError("id %r not found in module table" % module_id)
            scope = self.factory.createScope(self.row[S_FILE_ID], "module",
                                             module_id, rows[0])
        else:
            symbol_id = int(self.row[S_SCOPE_ID])
            rows = self.query(cu,
                 "SELECT * FROM symbol WHERE id=%d LIMIT 1" % symbol_id)
            if not rows:
                raise CIDBError("id %r not found in symbol table" % symbol_id)
            scope = self.factory.createScope(self.row[S_FILE_ID], "symbol",
                                             symbol_id, rows[0])
        return scope

    def getType(self):
        if self.table == "module":
            return "module"
        else:
            return self.row[S_TYPE]

    def query(self, cu, args):
        """Transparently cache SQL queries for this scope.
        
        Caching these on this Thingy instance will effective allow for
        proper invalidation of the cache: Thingy's are cached and invalidated
        by the ThingyFactory based on CIDB updating.
        """
        if isinstance(args, (str, unicode)):
            args = (args,)
        if not CACHING or args not in self.cache:
            cu.execute(*args)
            self.cache[args] = tuple(cu)
        return self.cache[args]

class Scope(Thingy):
    pass

class Import(Thingy):
    """Technically I said above that an 'Thingy' instance is limited to
    module or symbol entries. In general that is true, but for a short while
    in processing 'Import' objects can exist. They are alway resolved away
    to module or symbol entries, though.
    """
    def __str__(self):
        row = self.row
        module, symbol, alias = row[I_MODULE], row[I_SYMBOL], row[I_ALIAS]
        if alias and symbol:
            s = "from %(module)s import %(symbol)s as %(alias)s" % locals()
        elif alias:
            s = "import %(module)s as %(alias)s" % locals()
        elif symbol:
            s = "from %(module)s import %(symbol)s" % locals()
        else:
            s = "import %(module)s" % locals()
        return s

    def getType(self):
        return "import"


class ThingyFactory:
    """The may purpose in life for this class is cache management. It
    controls the creation and caching of all "Thingies" -- small objects
    used to represent row "hits" in the module, symbol and import tables
    of the CIDB.
    
    There is one instance of this factory (held by the controling
    codeintel.Manager instance) and that instance is passed around where
    necessary (generally as a variable named "factory").
    """
    def __init__(self, cache):
        # Note: the caches used here are keyed on the file_id/file_path to
        # enable efficient and correct cache invalidation on CIDB file
        # updates.

        # "thingyCache" is a mapping:
        #   <file_id>: { (<table>, <id>): <Thingy instances>,
        #                ... }
        self.thingyCache = cache.setdefault("thingyCache", {})

        # "scopeLineCache" is a mapping used to speed up the
        # "*ScopeForFileAndLine()" methods:
        #   <file_path>: (<scopingSymbolRows>, <scopingModuleRows>)
        # where <file_path> is the canonicalized file path.
        self.scopeLineCache = cache.setdefault("scopeLineCache", {})
        
        # "pythonFindModuleCache":
        #   {(<moduleName>, <submodule>, <cwd>): <findModule retval>,
        #    ...}
        # and "pythonFindModuleCacheIndex":
        #   {<file_id>: <key into pythonFindModuleCache>,
        #    ...}
        # are used to speed up PythonImportHandler.findModule:
        self.pythonFindModuleCache = cache.setdefault("pythonFindModuleCache", {})
        self.pythonFindModuleCacheIndex = cache.setdefault("pythonFindModuleCacheIndex", {})

        # "perlFindModuleCache":
        #   {(<moduleName>, <cwd>): <findModule retval>,
        #    ...}
        # and "perlFindModuleCacheIndex":
        #   {<file_id>: <key into perlFindModuleCache>,
        #    ...}
        # are used to speed up PerlImportHandler.findModule:
        self.perlFindModuleCache = cache.setdefault("perlFindModuleCache", {})
        self.perlFindModuleCacheIndex = cache.setdefault("perlFindModuleCacheIndex", {})
        
        # "builtInModuleCache":
        #   {<language>: <module Scope instance>}
        # is used to speed up Thingy.getParentScope. This mapping is a
        # weak reference dictionary to allow invalidation of the Scope
        # instance.
        self.builtInModuleCache = cache.setdefault("builtInModuleCache",
                                                   weakref.WeakValueDictionary())
        

    def invalidateCache(self, file_id, file_path):
        if file_id in self.thingyCache:
            del self.thingyCache[file_id]
        if file_path in self.scopeLineCache:
            del self.scopeLineCache[file_path]
        key = self.pythonFindModuleCacheIndex.get(file_id, None)
        if key is not None:
            del self.pythonFindModuleCache[key]
            del self.pythonFindModuleCacheIndex[file_id]
        key = self.perlFindModuleCacheIndex.get(file_id, None)
        if key is not None:
            del self.perlFindModuleCache[key]
            del self.perlFindModuleCacheIndex[file_id]

    def createThingy(self, file_id, table, id, row=None, cu=None):
        if CACHING:
            fileCache = self.thingyCache.setdefault(file_id, {})
            key = ("Thingy", table, id)
            if key not in fileCache:
                fileCache[key] = Thingy(self, file_id, table, id, row, cu)
            return fileCache[key]
        else:
            return Thingy(self, file_id, table, id, row, cu)
    def createScope(self, file_id, table, id, row=None, cu=None):
        if CACHING:
            fileCache = self.thingyCache.setdefault(file_id, {})
            key = ("Scope", table, id)
            if key not in fileCache:
                fileCache[key] = Scope(self, file_id, table, id, row, cu)
            return fileCache[key]
        else:
            return Scope(self, file_id, table, id, row, cu)
    def createImport(self, file_id, table, id, row=None, cu=None):
        if CACHING:
            fileCache = self.thingyCache.setdefault(file_id, {})
            key = ("Import", table, id)
            if key not in fileCache:
                fileCache[key] = Import(self, file_id, table, id, row, cu)
            return fileCache[key]
        else:
            return Import(self, file_id, table, id, row, cu)

    def createBuiltInModuleScope(self, cu, language):
        """Return a Scope object for the current language's built-in module.
        
        Built-in modules are represented in the CIDB with the name '*'.
        Returns None if there isn't one.
        """
        # We'll cache this, as will other caching, on the file_id. Note that
        # the file defining the built-in module is typically one of the
        # stdcix files (e.g. python.cix, perl.cix) that are loaded at
        # CIDB creation/preparation time. It is possible that another loaded
        # CIX could also define a "*" module, but the "first-one-wins"
        # semantics will mean those don't get used.
        if CACHING and language in self.builtInModuleCache:
            return self.builtInModuleCache[language]
        
        cu.execute("SELECT module.* FROM module, file, language "
                   "WHERE module.name='*' AND language.name=? "
                   "AND language.id=file.language_id "
                   "AND module.file_id=file.id LIMIT 1", (language,))
        for row in cu:
            scope = self.createScope(row[M_FILE_ID], "module", row[M_ID], row)
            break
        else:
            scope = None
        if CACHING:
            self.builtInModuleCache[language] = scope
        return scope

    def createScopeForFileAndLine(self, cu, path, line):
        """Return a Scope instance for the given position.
        
            "cu" is an open CIDB cursor.
            "file" is path to the file (not-canonicalized).
            "line" is the (1-based) line number at which to determine the
                scope.
        """
        cpath = canonicalizePath(path)
        scopingSymbolRows, scopingModuleRows = self.scopeLineCache.get(cpath, (None, None))
        
        # Return a symbol scope if the current line is in a
        # class/function/interface.
        # - Retrieve from, or update, the list of "scoping" symbols --
        #   ordered by line -- in the cache.
        if scopingSymbolRows is None:
            cu.execute("SELECT * FROM symbol "
                       "WHERE file_id=(SELECT id from file where compare_path=?) "
                       "AND type IN (?, ?, ?) ORDER BY line",
                       (cpath, ST_CLASS, ST_FUNCTION, ST_INTERFACE))
            scopingSymbolRows = tuple(cu)
        if CACHING:
            self.scopeLineCache[cpath] = (scopingSymbolRows, scopingModuleRows)
        # - XXX If this has perf impact use a binary search.
        symbolRow = None
        for row in scopingSymbolRows:
            if (row[S_LINE] <= line and
                (row[S_LINEEND] is None or line <= row[S_LINEEND])):
                symbolRow = row
            elif row[S_LINE] > line:
                break # we've already gone past the given "line"
        if symbolRow:
            scope = self.createScope(symbolRow[S_FILE_ID], "symbol",
                                     symbolRow[S_ID], symbolRow)
            return scope    

        # We are at module-level, return id of the appropriate module.
        if scopingModuleRows is None:
            cu.execute("SELECT * FROM module "
                       "WHERE file_id=(SELECT id from file where compare_path=?) "
                       "ORDER BY line", (cpath,))
            scopingModuleRows = tuple(cu)
        if CACHING:
            self.scopeLineCache[cpath] = (scopingSymbolRows, scopingModuleRows)
        if not scopingModuleRows:
            raise CodeIntelError("there are no CIDB module entries for "
                                 "file '%s'" % path)
        moduleRow = None
        for row in scopingModuleRows:
            if row[M_LINE] <= line:
                moduleRow = row
            elif row[M_LINE] > line:
                break # we've gone past the given "line"
        if not moduleRow:
            #XXX Why is this being hit so frequently? Figure it out.
            raise CodeIntelError("unexpectedly, no module rows for file "
                                 "'%s' correspond to line %d"
                                 % (cpath, line))
        scope = self.createScope(moduleRow[M_FILE_ID], "module",
                                 moduleRow[M_ID], moduleRow)
        return scope

    def createModuleScopeForFileAndLine(self, cu, path, line):
        """Return a Scope instance representing the current module."""
        cpath = canonicalizePath(path)
        scopingSymbolRows, scopingModuleRows = self.scopeLineCache.get(cpath, (None, None))
        if scopingModuleRows is None:
            cu.execute("SELECT * FROM module "
                       "WHERE file_id=(SELECT id from file where compare_path=?) "
                       "ORDER BY line", (cpath,))
            scopingModuleRows = tuple(cu)
        if CACHING:
            self.scopeLineCache[cpath] = (scopingSymbolRows, scopingModuleRows)
        if not scopingModuleRows:
            raise CodeIntelError("there are no CIDB module entries for "
                                 "file '%s'" % path)
        moduleRow = None
        for row in scopingModuleRows:
            if row[M_LINE] <= line:
                moduleRow = row
            elif row[M_LINE] > line:
                break # we've gone past the given "line"
        if not moduleRow:
            #XXX Why is this being hit so frequently? Figure it out.
            raise CodeIntelError("unexpectedly, no module rows for file "
                                 "'%s' correspond to line %d"
                                 % (cpath, line))
        scope = self.createScope(moduleRow[M_FILE_ID], "module",
                                 moduleRow[M_ID], moduleRow)
        return scope



class Context:
    # May want to cache data for the current file rather than starting
    # from zero for each getMembers/getCallTips in that file.
    # c.f. Memoize: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52201
    def __init__(self, manager, language, path, line):
        """Create a CITDL evaluation Context.
        
            "manager" is the controlling codeintel.Manager instance.
            "language" is the file language in which to evaluate.
            "path" is the path to the file.
            "line" is the (1-based) trigger line. The path and line together
                typically determine the scope of evaluation.
            "cache" a persistent (on the lifetime of the Manager) cache
                dictionary.
        """
        log.debug("   Context(manager, language=%r, path=%r, "
                  "line=%r)", language, path, line)
        self.manager = manager
        self.factory = manager.factory # a ThingyFactory instance to use
        self.language = language
        self.path = path
        self.cwd = os.path.dirname(path)
        self.line = line
        self.cache = manager.cache
        self.file_id = None # determined lazily
        self.module_id = None # determined lazily
        self.notes = []

        # Sentinels for trapping infinite recursion:
        # An internal list of objects on whom members have been retrieved.
        # This is used to guard against infinite loops (see
        # test/test_eval_citdl/infinity.py). It is reset for each "stage"
        # of the top-level evaluation. It was quite delicate to get this
        # to trap infinite loops but still allow proper evaluation.
        # c.f. test_citdl.PythonTestCase.test_infinity
        #      test_citdl.PerlTestCase.test_usestdmodules_LWP
        self.__hitThingies = None
        self.__typeInferenceHits = {}

    def getMembers(self, cu, expr,
                   scopeFileId=0, scopeTable=None, scopeId=0):
        """Get the members for the given CITDL expression.
        
            "cu" is an open CIDB cursor
            "expr" is the CITDL expression
            "scopeFileId", "scopeTable" and "scopeId" (optional) can be
                passed in to express the scope more accurately than "path"
                and "line" -- if, say, the caller is adjusting for recent
                edits not in the database.
        
        Returns a list of 2-tuples: (<type-name>, <member-name>)
        The <type-name> may be None, to indicate that the type is unknown.
        """
        self.log("getMembers for '%s' at %s:%d", expr,
                 os.path.basename(self.path), self.line)
        if scopeTable:
            scope = self.factory.createScope(scopeFileId, scopeTable,
                                             scopeId, cu=cu)
        else:
            scope = self.factory.createScopeForFileAndLine(cu, self.path,
                                                           self.line)
        if self.language == "Perl":
            self.perlPkgInfo = {}
        else:
            self.perlPkgInfo = None
        members = self._getMembers(cu, expr, scope, _toplevel=True)
        self.log("    members: %r", members)
        return members

    def getCallTips(self, cu, expr,
                    scopeFileId=0, scopeTable=None, scopeId=0):
        """Get the members for the given CITDL expression.
        
            "cu" is an open CIDB cursor
            "expr" is the CITDL expression
            "scopeFileId", "scopeTable" and "scopeId" (optional) can be
                passed in to express the scope more accurately than "path"
                and "line" -- if, say, the caller is adjusting for recent
                edits not in the database.
        """
        self.log("getCallTips for '%s' at %s:%d", expr,
                 os.path.basename(self.path), self.line)
        if scopeTable:
            scope = self.factory.createScope(scopeFileId, scopeTable,
                                             scopeId, cu=cu)
        else:
            scope = self.factory.createScopeForFileAndLine(cu, self.path,
                                                           self.line)
        if self.language == "Perl":
            self.perlPkgInfo = {}
        else:
            self.perlPkgInfo = None
        calltips = self._getCallTips(cu, expr, scope, _toplevel=True)
        self.log("    calltips: %r", calltips)
        return calltips

    def log(self, msg, *args):
        self.notes.append(msg % args)
        log.info(msg, *args)
    def warn(self, msg, *args):
        self.notes.append(msg % args)
        log.info(msg, *args) # don't want to alarm user's with warn-level log output
    def error(self, msg, *args):
        self.notes.append(msg % args)
        log.error(msg, *args)

    def _parseScope(self, cu, expr, scope=None):
        """Determine and return the appropriate scope.

        The scope is determined like this:
        - the "expr" expression can specify it, e.g. in
          "sys.version_info Foo.bar" the scope is "Foo.bar"
        - otherwise, if "scope" is specified, then that is it
        - otherwise, module-level scope is implied.

        This returns (1) the CITDL <symbol-ref> part, with any possible
        <start-scope> part removed; and, (2) the scope -- a Scope instance.
        """
        if ' ' in expr:
            try:
                symbolRef, scopeStr = expr.split()
            except ValueError, ex:
                raise CITDLError("invalid CITDL expression: can only have "
                                 "two parts: %r" % expr)
            startScope = self._resolveStartScope(cu, scopeStr)
        elif scope is not None:
            symbolRef = expr
            startScope = scope
        else:
            # Get the module-level scope.
            symbolRef = expr
            startScope = self.factory.createModuleScopeForFileAndLine(
                cu, self.path, self.line)

        retval = (symbolRef, startScope)
        log.debug("   Context._parseScope(expr=%r, scope=%r): return %r",
                  expr, scope, retval)
        return retval

    def _resolveStartScope(self, cu, scopeStr):
        """Resolve the given <start-scope> string part of a CITDL expr.

        This is a simpler expression than the <symbol-ref> part of a full
        CITDL expression: '()' tokens are illegal and only scoping symbols
        (classes, interfaces, functions) and modules are valid. I.e. no
        imports and no variables/arguments.
        
        Returns a Scope instance for the resolved scope or raises a
        CodeIntelError.
        """
        #XXX Don't bother checking this assertion for now.
        #if '(' in scopeStr or ')' in scopeStr:
        #    raise CITDLError("illegal CITDL syntax: '()' not allowed in "
        #                     "<start-scope> part: '%s'" % scopeStr)

        # <start-scope> parts are always based at module-scope
        scope = self.factory.createModuleScopeForFileAndLine(
            cu, self.path, self.line)
        for token in scopeStr.split('.'):
            if scope.table == "module":
                rows = scope.query(cu,
                    ("SELECT * FROM symbol "
                     " WHERE module_id=? AND NOT scope_id "
                     "   AND name=? AND type IN (?, ?, ?) LIMIT 1",
                     (scope.id, token, ST_CLASS, ST_FUNCTION, ST_INTERFACE)))
            else: # scope.table == "symbol"
                rows = scope.query(cu,
                    ("SELECT * FROM symbol "
                     " WHERE scope_id=? AND name=? "
                     "   AND type IN (?, ?, ?) LIMIT 1",
                     (scope.id, token, ST_CLASS, ST_FUNCTION, ST_INTERFACE)))
            if not rows:
                raise CodeIntelError("could not resolve '%s': could not "
                                     "find '%s' on %r"
                                     % (scopeStr, token, scope))
            row = rows[0]
            scope = self.factory.createScope(row[S_FILE_ID], "symbol",
                                             row[S_ID], row)

        self.log("    start scope '%s' resolves to %r", scopeStr, scope)
        return scope

    def _eval(self, cu, expr, scope=None, _toplevel=False):
        """Evaluate the given CITDL expression at the given scope.

            "cu" is an open CIDB cursor.
            "expr" is the CITDL expression to evaluate.
            "scope" (optional) can be used to define a specific scope at
                which to begin evaluation.
        
        A CITDL expression is of the following form:
          <symbol-ref>[ <start-scope>]
        where,
          <symbol-ref> is a chain of identifiers, '.' for attribute access,
              and '()' for calling, e.g. "os.environ.get().split()"
          <start-scope> is a chain "scoping" identifiers and '.' for
              attribute access. "Scoping" here means that only symbols that
              create new scopes are relevant, i.e. classes, functions and
              interfaces but not arguments and variables. A start scope
              is always in the current module so a module reference is
              not relevant either.
        """
        log.debug("   Context._eval(expr=%r, scope=%r)", expr, scope)
        symbolRef, startScope = self._parseScope(cu, expr, scope)

        # Try to resolve the CITDL expression in the current scope,
        # then try to the parent scope, ... then try the module-level scope,
        # (XXX: then try to "builtins-scope" (a.k.a. the "*" module)).
        scope = startScope
        while 1:
            hits = self._evalAtScope(cu, symbolRef, scope, _toplevel=_toplevel)
            if hits:
                # Guard against an infinite loop.
                newhits = []
                for hit in hits:
                    guid = hit.guid()
                    if guid in self.__hitThingies:
                        log.info("    possible infinite loop on %r: "
                                 "dropping it", hit)
                    else:
                        self.__hitThingies[guid] = 1
                        newhits.append(hit)
                hits = newhits
                break
            else:
                scope = scope.getParentScope(cu, self.language)
            if not scope:
                hits = []
                break

        return hits

    def _evalAtScope(self, cu, expr, scope, _toplevel=False):
        """Attempt to resolve the given CITDL symbol reference at the
        given scope.

            "cu" is an open CIDB cursor.
            "expr" is the <symbol-ref> part of a CITDL expression.
            "scope" is the Scope on which to search.

        Returns a list of Thingy instances (one for each hit) or an empty list
        if there were no hits at this scope. It can raise a CodeIntelError for
        limitations or other errors.
        """
        log.debug("   Context._evalAtScope(expr=%r, scope=%r)", expr, scope)
        hits = [scope]
        tokenizer = re.compile("(\.|\(\))")
        isFirstAttrToken = 1
        for token in tokenizer.split(expr):
            if _toplevel:
                log.debug("   ------ substage '%s' (%r) ------", token, scope)
                self.__hitThingies = {}
            
            if token == '.':
                continue
            elif token == '()':
                raise CodeIntelError("cannot yet evaluate CITDL expressions "
                                     "using '()': '%s'" % expr)
            else:
                attrHits = []
                for hit in hits:
                    attrHits += self._resolveGetattr(cu, hit, token,
                                                     isFirstAttrToken)
                    isFirstAttrToken = 0
                hits = attrHits
        return hits

    def _resolveImport(self, cu, impobj, name, first=0):
        """Resolve the given import to a symbol or module object.

            "obj" is an Import instance, i.e. an import table instance
            "name" is the name of the symbol or module to find on "obj".
            "first" is a boolean indicating if this is the first symbol name
                in the CITDL expression (it matters for functions).

        Returns a list of possible hits for the import lookup. If the module
        or specifically named symbol (i.e. not '*') cannot be found in the
        database, an error is raised. If a symbol cannot be found in a
        '*'-type import, that is fine, the empty list is returned.
        """
        log.debug("   Context._resolveImport(impobj=%r, name=%r, first=%r)",
                  impobj, name, first)
        hits = []

        moduleName = impobj.row[I_MODULE]
        symbol = impobj.row[I_SYMBOL]
        alias = impobj.row[I_ALIAS]

        handler = self.manager.getImportHandler(self.language)
        module, kind = handler.findModule(cu, self.factory, moduleName,
                                          symbol, cwd=self.cwd)
        log.debug("   Context._resolveImport: moduleName=%r, symbol=%r, "
                  "alias=%r: %r", moduleName, symbol, alias, module)
        module_id = module.id
        
        # Import resolution is significantly different for Perl.
        if self.language == "Perl":
            # All Perl use-statement types effectively import the packages
            # defined (or touched) in that module. I.e., after any of
            #   use Data::Dumper ();
            #   use Data::Dumper;
            #   use Data::Dumper qw(Dumper);
            #   use Data::Dumper qw(:tag);
            # one can access fully-qualified Data::Dumper::* members.
            # Likewise after "require" statements.
            rows = module.query(cu,
                ("SELECT * FROM symbol WHERE module_id=? AND NOT scope_id "
                 "AND type=? AND name=?", (module_id, ST_CLASS, name)))
            for row in rows:
                hits = [ self.factory.createThingy(row[S_FILE_ID], "symbol",
                                                   row[S_ID], row) ]

            # Get the ID of the package of the same name as the module
            # which is presumed to be defined in this module. This is
            # used for handling most "use"-statement types below.
            package_id = None
            if symbol:
                rows = module.query(cu,
                    ("SELECT id FROM symbol "
                     "WHERE module_id=? AND type=? AND name=? LIMIT 1",
                     (module_id, ST_CLASS, moduleName)))
                if rows:
                    package_id = rows[0][0]
                else:
                    log.warn("could not find '%s' package in '%s' Perl module",
                             moduleName, moduleName)

            if not symbol:
                # Covers:   use Foo ();
                pass
            elif package_id is None:
                # Cannot handle other "use"-statement types without this
                # package definition.
                #XXX This is probably not completely correct!
                pass
            elif symbol == "*":
                # Covers:   use Foo;
                rows = module.query(cu,
                    ("SELECT * FROM symbol "
                     "WHERE module_id=? AND scope_id=? "
                     "AND name=? "
                     "AND attributes LIKE '%exported%'",
                     (module_id, package_id, name)))
                for row in rows:
                    hits.append( self.factory.createThingy(
                                    row[S_FILE_ID], "symbol", row[S_ID], row) )
            elif symbol == "**":
                # Covers:   use Foo qw(:tag);
                # I.e. our hack fallback for Perl's %EXPORT_TAGS.
                rows = module.query(cu,
                    ("SELECT * FROM symbol "
                     "WHERE module_id=? AND scope_id=? "
                     "AND name=? "
                     # either __exported__ or __exportable__
                     "AND attributes LIKE '%export%'",
                     (module_id, package_id, name)))
                for row in rows:
                    hits.append( self.factory.createThingy(row[S_FILE_ID],
                                    "symbol", row[S_ID], row) )
            elif symbol == name:
                # Covers:   use Foo qw(bar baz);
                # Search all symbols in the appropriate Perl package in this
                # module file.
                rows = module.query(cu,
                    ("SELECT * FROM symbol "
                     "WHERE module_id=? AND scope_id=? AND name=?",
                     (module_id, package_id, name)))
                for row in rows:
                    hits.append( self.factory.createThingy(row[S_FILE_ID],
                                    "symbol", row[S_ID], row) )

            if self.perlPkgInfo is not None:
                # This is our Perl hack to determine which packages are
                # available after Perl module imports. For example, after
                # "use LWP;" package HTTP::Request should be available.
                self.log("    accumulate package info in '%s'", moduleName)
                self._accumulatePerlPkgInfo(cu, module_id, moduleName)
                if log.isEnabledFor(logging.DEBUG):
                    log.debug("   loaded Perl packages after '%s':", moduleName)
                    for pkg in self.perlPkgInfo:
                        log.debug("       %s", pkg)

                # If we haven't found anything, perhaps the name is a package
                # that has been loaded by recursive imports.
                if not hits and name in self.perlPkgInfo:
                    row = self.perlPkgInfo[name]
                    hits = [ self.factory.createThingy(row[S_FILE_ID],
                                "symbol", row[S_ID], row) ]

        # Import resolution for non-Perl (just Python now) imports follows.
        elif (kind == "module" and not symbol) or kind == "submodule":
            # Covers:   import mod
            #           import mod as alias
            #           from pkg import mod
            #           from pkg import mod as alias
            hits = [ module ]
        else:
            # Covers:   from mod import symbol
            #           from mod import symbol as alias
            #           from mod import *
            # Have to lookup the appropriate symbol on the found module.
            if alias:
                symbolName = symbol
            else:
                symbolName = name
            # Search all top-level symbols defined in this module with
            # the appropriate name.
            rows = module.query(cu,
                ("SELECT * FROM symbol "
                 "WHERE module_id=? AND NOT scope_id AND name=?",
                 (module_id, symbolName)))
            for row in rows:
                hits.append( self.factory.createThingy(row[S_FILE_ID], "symbol", row[S_ID], row) )
            # Also add recursive imports. I.e. if you imported a symbol
            # from module Foo and that module itself imported the symbol
            # from module Bar, this will get it.
            rows = module.query(cu, 
                ("SELECT * FROM import "
                 "WHERE module_id=? AND NOT scope_id "
                 "AND (alias=? OR "
                 "     (NOT alias AND "
                 "      (symbol=? OR "
                 "       (NOT symbol AND module=?))))",
                 (module_id, symbolName, symbolName, symbolName)))
            for irow in rows:
                iobj = self.factory.createImport(row[I_FILE_ID], "import",
                                                 irow[I_ID], irow)
                ihits = self._resolveImport(cu, iobj, symbolName, first)
                hits += ihits

        self.log("    does '%s' import '%s'? %s", impobj, name,
                 hits and "yes" or "no")
        if log.isEnabledFor(logging.INFO) and hits:
            for hit in hits:
                self.log("        %r", hit)
        return hits

    def _accumulatePerlPkgInfo(self, cu, module_id, module_name, _level=0):
        #log.debug("   _accumulatePerlPkgInfo(module_id=%r, module_name=%r, _level=%r)",
        #          module_id, module_name, _level)

        # Use cached data, if available.
        #
        # "perlPkgInfoCache" is a dict mapping:
        #   <perl module name>: (<package data>
        #                        <import data>)
        perlPkgInfoCache = self.cache.setdefault("perlPkgInfoCache", {})
        pkgData, importData = perlPkgInfoCache.get(module_name, (None, None))
        
        # Add all packages defined in this module.
        if pkgData is None:
            pkgData = {}
            cu.execute("SELECT * FROM symbol "
                       "WHERE module_id=? AND type=?",
                       (module_id, ST_CLASS))
            for row in cu:
                pkgData[row[S_NAME]] = row
            perlPkgInfoCache[module_name] = (pkgData, None)
        self.perlPkgInfo.update(pkgData)

        # Optmization: only look two levels deep.
        if _level > 1:
            return

        # Recurse into imports in this module.
        if importData is None:
            # Cheat: Instead of looking for the file for this import we just
            # grab the first scanned module with the right name.
            cu.execute("SELECT module.id,module.name FROM import,module "
                       "WHERE import.module_id=%d AND module.name=import.module"
                       % module_id)
            importData = {} # avoid duplicates
            for row in cu:
                importData[tuple(row)] = 1
            perlPkgInfoCache[module_name] = (pkgData, importData)
        for (modid, modname) in importData:
            self._accumulatePerlPkgInfo(cu, modid, modname, _level+1)

    def _getSymbolTypeInferences(self, cu, symbol, symbolType):
        symbol_id = symbol.id
        if symbol_id in self.__typeInferenceHits:
            log.info("    possible infinite loop on %r type inferences: "
                     "skipping", symbol)
            return []
        else:
            self.__typeInferenceHits[symbol_id] = 1

        rows = symbol.query(cu, ("SELECT citdl FROM type WHERE symbol_id=%d" % symbol.id))
        typeInferences = [row[0] for row in rows]
        self.log("    type inferences for %s symbol %d: %r",
                 symbolType2Name(symbolType), symbol.id, typeInferences)
        return typeInferences

    def _resolveGetattr(self, cu, obj, name, first=0):
        """Resolve the given attribute/symbol lookup.

            "obj" is an Thingy instance, i.e. a module or symbol row
                reference. It is the object on which to resolve "name"
                attribute lookup.
            "name" is the name of the symbol to find on "obj".
            "first" is a boolean indicating if this is the first symbol name
                in the CITDL expression (it matters for functions).

        Returns a list of possible hits for the attr lookup or the empty
        list if it could not be resolved.
        """
        log.debug("   Context._resolveGetattr(obj=%r, name=%r, first=%r)",
                  obj, name, first)
        hits = {} # use dict to eliminate dupes, maps Thingy.guid() to Thingy
        if obj.table == "module":
            symbolType = None
        else:
            symbolType = int(obj.row[S_TYPE])

        if symbolType in (ST_VARIABLE, ST_ARGUMENT):
            typeInferences = self._getSymbolTypeInferences(cu, obj, symbolType)
            for expr in typeInferences:
                for o in self._eval(cu, expr):
                    for h in self._resolveGetattr(cu, o, name, first):
                        hits[h.guid()] = h
        elif not first and symbolType == ST_FUNCTION:
            # Function arguments and internal variables are NOT attributes of
            # the function. Therefore, these should only resolve if they are
            # the first part of a CITDL expression. E.g., given this function:
            #   def foo(arg1, arg2):
            #       var1 = 42
            #       #...
            # "arg1" and "var1" _in_ the function should resolve, but
            # "foo.var1" and "foo.arg1" should NOT.
            pass
        else:
            # Check imports for a match.
            log.debug("   Context._resolveGetattr: check imports for "
                      "a match of '%s'", name)

            if obj.table == "module":
                query = ("SELECT * FROM import "
                         "WHERE module_id=%d AND NOT scope_id" % obj.id)
            else:
                query = "SELECT * FROM import WHERE scope_id=%d" % obj.id
            rows = obj.query(cu, query) # select imports at this scope
            for irow in rows:
                # Only descend into an import if it looks like it is possible
                # that is defines "name". Note that this is *always* complex
                # for any "use" statement in Perl (any packages in that
                # module are essentially imported) so we always need to
                # descend into them.
                if self.language == "Perl" \
                   or (irow[I_ALIAS] or irow[I_SYMBOL] or irow[I_MODULE]) in (name, "*", "**"):
                    impobj = self.factory.createImport(
                        irow[I_FILE_ID], "import", irow[I_ID], irow)
                    try:
                        for o in self._resolveImport(cu, impobj, name):
                            hits[o.guid()] = o
                    except CodeIntelError, ex:
                        self.warn("    error resolving %r: ignoring",
                                  impobj)

            # Check locally defined symbols for a match.
            if obj.table == "module":
                query = ("SELECT * FROM symbol "
                         "WHERE module_id=? AND NOT scope_id AND name=?",
                         (obj.id, name))
            else:
                query = ("SELECT * FROM symbol WHERE scope_id=? AND name=?",
                         (obj.id, name))
            gotLocalHit = False
            for row in obj.query(cu, query): # select matching symbols at this scope
                o = self.factory.createThingy(row[S_FILE_ID], "symbol",
                                              row[S_ID], row)
                hits[o.guid()] = o
                gotLocalHit = True

            # Regarding the "not isinstance(obj, Scope)" here: When doing
            # getattr on a class it matters if we are using the class as a
            # class _object_ (i.e. a Thingy instance) or as just a lexical
            # scope (i.e. a Scope instance). In the former we want to dive
            # into base classes (and interfaces). In the latter we don't.
            # This is both (1) correct:
            #   class Foo:
            #       a = "blah"
            #   class Bar(Foo):
            #       # 'a' should NOT resolve here, e.g. "a.upper() should fail"
            #   Bar.a  # 'a' SHOULD resolve here
            # and (2) required to avoid infinite recursion in Perl because
            # base package (aka class) references in Perl are resolved at
            # that package's scope (see Thingy.getBaseScope()).
            if not isinstance(obj, Scope) and symbolType == ST_CLASS and not gotLocalHit:
                # Check base classes and interfaces (if relevant).
                bases = obj.query(cu, "SELECT id, name FROM base WHERE symbol_id=%d" % obj.id)
                if bases:
                    self.log("    bases for type:%s symbol %d: %r",
                             symbolType, obj.id, [b[1] for b in bases])
                    baseScope = obj.getBaseScope(cu, self.language)
                    for baseId, baseName in bases:
                        btrows = obj.query(cu,
                            "SELECT citdl FROM type WHERE base_id=%d" % baseId)
                        baseTypeInferences = [btrow[0] for btrow in btrows]
                        for expr in baseTypeInferences:
                            for o in self._eval(cu, expr, baseScope):
                                for h in self._resolveGetattr(cu, o, name,
                                                              first):
                                    hits[h.guid()] = h

        hits = hits.values()
        self.log("    is '%s' accessible on %r? %s", name, obj,
                 hits and "yes" or "no")
        if log.isEnabledFor(logging.INFO) and hits:
            for hit in hits:
                self.log("        %r", hit)
        return hits
    
    def _getSymbolMembers(self, cu, symbol):
        members = []
        symbolType = int(symbol.row[S_TYPE])
        if symbolType == ST_FUNCTION:
            # Function arguments and internal variables should NOT be shown
            # in an autocomplete list for the function. Those would be
            # function attributes, if we supported them -- which we currently
            # don't. (Do many languages support/use them -- Python does, but
            # their use is rare, I think).
            pass
        elif symbolType in (ST_CLASS, ST_INTERFACE):
            rows = symbol.query(cu,
                ("SELECT name, type FROM symbol WHERE scope_id=%d AND "
                 " (attributes ISNULL OR attributes NOT LIKE '%%__local__%%')"
                 % symbol.id))
            members = {}
            for name, type in rows:
                #XXX Will want to pass 'type' eventually to
                #    put proper icons in autocomplete UI.
                members[(symbolType2Name(type), name)] = 1
            # Add base class members.
            bases = symbol.query(cu, "SELECT id, name FROM base WHERE symbol_id=%d" % symbol.id)
            if bases:
                self.log("    bases for type:%s symbol %d: %r",
                         symbolType, symbol.id, [b[1] for b in bases])
                baseScope = symbol.getBaseScope(cu, self.language)
                
                for baseId, baseName in bases:
                    if self.language == "Perl":
                        # Special case: skip "Exporter" base class members in
                        # Perl. These are not desired 99% of the time.
                        btrows = symbol.query(cu,
                            "SELECT citdl FROM type WHERE base_id=%d "
                            "AND NOT (citdl='Exporter' OR citdl='Exporter%%')" % baseId)
                    else:
                        btrows = symbol.query(cu,
                            "SELECT citdl FROM type WHERE base_id=%d" % baseId)
                    baseTypeInferences = [btrow[0] for btrow in btrows]
                    self.log("    type inferences for base '%s' (%d): %r",
                             baseName, baseId, baseTypeInferences)
                    baseMembers = []
                    for expr in baseTypeInferences:
                        try:
                            baseMembers += self._getMembers(cu, expr,
                                                            baseScope)
                        except CodeIntelError, ex:
                            self.warn("    error getting members from "
                                      "base '%s': %s: ignoring", expr, ex)
                    self.log("    base '%s' members: %r", baseName, baseMembers)
                    for baseMember in baseMembers:
                        if self.language == "Perl" and baseMember[0] == "class":
                            # Perl packages (represented as classes) should
                            # NOT inherit the sub-packages of their base
                            # classes. E.g. a package that @ISA=qw(Exporter)
                            # should not always have the "Heavy" (from
                            # Exporter::Heavy) as a member.
                            continue
                        members[baseMember] = 1
            members = members.keys()
        elif symbolType in (ST_VARIABLE, ST_ARGUMENT):
            typeInferences = self._getSymbolTypeInferences(cu, symbol, symbolType)
            if typeInferences:
                scope = symbol.getParentScope(cu, self.language)
            for expr in typeInferences:
                members += self._getMembers(cu, expr, scope)
        else:
            raise CodeIntelError("cannot get members from CITDL that "
                                 "resolves to a '%(type)s' symbol: "
                                 "%(name)s (id=%(id)d)" % symbol.row)
        return members

    def _getModuleMembers(self, cu, module):
        members = []
        rows = module.query(cu, "SELECT name, type FROM symbol "
                                "WHERE module_id=%d AND NOT scope_id" % module.id)
        for (name, type) in rows:
            members.append((symbolType2Name(type), name))

        rows = module.query(cu,
            "SELECT * FROM import WHERE module_id=%d AND NOT scope_id" % module.id)
        for irow in rows:
            iobj = self.factory.createImport(irow[I_FILE_ID], "import",
                                             irow[I_ID], irow)
            members += self._getImportedMembers(cu, iobj)

        return members

    def _getImportedMembers(self, cu, impobj):
        # Returning "None" for the type indicates "I don't know what the
        # member type is". We *could* do extra processing to track down
        # these members and get their types. Not bothering because of the
        # extra time it would take.
        row = impobj.row
        module, symbol, alias = row[I_MODULE], row[I_SYMBOL], row[I_ALIAS]
        if alias:
            members = [(not symbol and "module" or None, alias)]
        elif symbol == "*":
            log.info("    warning: don't yet know how to get '*' import: "
                     "'from %s import *'", module)
            members = []
        elif symbol:
            members = [(None, symbol)]
        elif module:
            members = [("module", module)]
        return members

    def _getMembers(self, cu, expr, scope=None, _toplevel=False):
        """Get the members for the given CITDL expression and scope.
        
            "cu" is an open CIDB cursor.
            "expr" is a CITDL expression.
            "scope" (optional) is a Scope instance specifying the start
                scope. Not that a scope specified in "expr" might still
                override this (see comment for _eval()).
        """
        log.debug("   _getMembers(expr=%r, scope=%r)", expr, scope)
        if _toplevel:
            log.debug("   ====== stage 1: toplevel eval ======")
            self.__hitThingies = {}
        hits = self._eval(cu, expr, scope)
        if _toplevel:
            log.debug("   ====== stage 2: toplevel use eval'd hits ======")
            self.__hitThingies = {}

        members = Set() # use a Set to ensure no duplicates
        for hit in hits:
            if hit.table == "module":
                members.union_update( self._getModuleMembers(cu, hit) )
            elif hit.table == "symbol":
                members.union_update( self._getSymbolMembers(cu, hit) )
            else:
                raise CodeIntelError("cannot get members from CITDL that "
                                     "resolves to a '%s': %r"
                                     % (hit.table, expr))

        log.debug("   _getMembers: perlPkgInfo=%r", self.perlPkgInfo)
        if self.perlPkgInfo:
            # Efficiently extract the sub-package completions -- from
            # available Perl packages -- for the CITDL expression.
            perlPkgs = self.perlPkgInfo.keys()
            perlPkgs.sort()
            pkgPrefix = expr.split()[0]+"::"
            index = bisect.bisect_left(perlPkgs, pkgPrefix)
            for pkg in perlPkgs[index:]:
                if not pkg.startswith(pkgPrefix): break
                subpkg = pkg[len(pkgPrefix):].split("::")[0]
                members.add(("class", subpkg))
            
        return list(members)

    def _getSymbolCallTips(self, cu, symbol):
        from codeintel.util import LINE_LIMIT
        symbolType = int(symbol.row[S_TYPE])
        calltips = []

        if symbolType == ST_FUNCTION:
            # See "Determining a Function CallTip" in the spec for a
            # discussion of this algorithm.
            signature = symbol.row[S_SIGNATURE]
            doc = symbol.row[S_DOC]
            ctlines = []
            if not signature:
                name = symbol.row[S_NAME]
                if self.language in ("Tcl", ):
                    ctlines = [name + " ..."]
                else:
                    ctlines = [name + "(...)"]
            else:
                ctlines = signature.split('\n')
            if doc:
                ctlines += doc.split('\n')[:LINE_LIMIT-len(ctlines)]
            calltips.append('\n'.join(ctlines))
            
        elif symbolType == ST_CLASS:
            # If the class has a defined signature then use that.
            className = symbol.row[S_NAME]
            classSig = symbol.row[S_SIGNATURE]
            classDoc = symbol.row[S_DOC]
            if classSig:
                ctlines = classSig.split('\n')
                if classDoc:
                    ctlines += classDoc.split('\n')[:LINE_LIMIT-len(ctlines)]
                calltips.append('\n'.join(ctlines))
            
            # Alternatively we use calltip information on the class'
            # constructor. In the ideal world we would:
            # - find all constructor methods (including those in base
            #   classes);
            # - eliminate those that are overridden (based on
            #   language-specific details on when this occurs, e.g.: method
            #   name match in Python, signature match in C++); and
            # - return them all
            # For reasons of performance:
            # - Because we currently only show one calltip in the UI we can
            #   either only return one or not bother culling overridden
            #   constructor -- whichever is easier.
            else:
                #XXX We are assuming that we are not getting false positives
                #    in the attribute comparison (given that '_' is a
                #    wildcard char for SQL's LIKE keyword).
                ctorSymbols = symbol.query(cu,
                    ("SELECT * FROM symbol WHERE scope_id=? "
                     "   AND type=? AND attributes LIKE '%__ctor__%'",
                     (symbol.id, ST_FUNCTION)))
                if not ctorSymbols:
                    bases = symbol.query(cu, "SELECT id, name FROM base WHERE symbol_id=%d" % symbol.id)
                    if bases:
                        self.log("    bases for type:%s symbol %d: %r",
                                 symbolType, symbol.id, [b[1] for b in bases])
                        baseScope = symbol.getBaseScope(cu, self.language)
                        for baseId, baseName in bases:
                            btrows = symbol.query(cu, "SELECT citdl FROM type WHERE base_id=%d" % baseId)
                            baseTypeInferences = [btrow[0] for btrow in btrows]
                            for expr in baseTypeInferences:
                                for hit in self._eval(cu, expr, baseScope):
                                    ctorSymbols += hit.query(cu,
                                        ("SELECT * FROM symbol "
                                         " WHERE scope_id=? "
                                         "   AND type=? "
                                         "   AND attributes LIKE '%__ctor__%'",
                                         (hit.id, ST_FUNCTION)))
                                    if ctorSymbols: break  # Optimization
                                if ctorSymbols: break  # Optimization
                            if ctorSymbols: break  # Optimization
    
                if self.language in ("Tcl",):
                    fallbackSig = className + " ..."
                else:
                    fallbackSig = className + "(...)"
                if ctorSymbols:
                    if len(ctorSymbols) > 1: # Optimization: drop multiple ctors
                        log.debug("   ignoring all but first constructor "
                                  "for class %s", className)
                    calltip = ctorSymbols[0][S_SIGNATURE] or fallbackSig
                    doc = ctorSymbols[0][S_DOC] or classDoc
                    if doc:
                        ctlines = calltip.split('\n') + doc.split('\n', LINE_LIMIT)
                        ctlines = ctlines[:LINE_LIMIT]
                        calltip = '\n'.join(ctlines)
                else:
                    calltip = fallbackSig
                calltips.append(calltip)

        elif symbolType in (ST_ARGUMENT, ST_VARIABLE):
            typeInferences = self._getSymbolTypeInferences(cu, symbol, symbolType)
            if typeInferences:
                scope = self.factory.createScope(symbol.file_id, "module",
                                                 symbol.row[S_MODULE_ID], cu=cu)
            for expr in typeInferences:
                calltips += self._getCallTips(cu, expr, scope)

        else:
            raise CodeIntelError("cannot get calltips from CITDL that "
                                 "resolves to a '%(type)s' symbol: "
                                 "%(name)s (id=%(id)d)" % symbol.row)
        return calltips

    def _getCallTips(self, cu, expr, scope=None, _toplevel=False):
        """Get the calltips for the given CITDL expression and scope.
        
            "cu" is an open CIDB cursor.
            "expr" is a CITDL expression.
            "scope" (optional) is a Scope instance specifying the start
                scope. Not that a scope specified in "expr" might still
                override this (see comment for _eval()).
        """
        log.debug("   Context._getCallTips(expr=%r, scope=%r)", expr, scope)
        calltips = []
        if _toplevel:
            log.debug("   ====== stage 1: toplevel eval ======")
            self.__hitThingies = {}
        hits = self._eval(cu, expr, scope, _toplevel=_toplevel)
        if _toplevel:
            log.debug("   ====== stage 2: toplevel use eval'd hits ======")
            self.__hitThingies = {}

        for hit in hits:
            if hit.table == "module":
                pass # modules are not callable
            elif hit.table == "symbol":
                #XXX Should trap and only pass through error here if no
                #    calltip for *any* of the hits could be found.
                calltips += self._getSymbolCallTips(cu, hit)
            else:
                raise CodeIntelError("cannot get calltips from CITDL that "
                                     "resolves to a '%s': %r"
                                     % (hit.table, expr))

        if len(calltips) > 1:
            # Apply a rudimentary order: preferred first.
            calltips.sort(_preferredCallTipsFirst)

        log.debug("   Context._getCallTips: calltips=%s", calltips)
        return calltips



#---- mainline self-test

if __name__ == '__main__':
    def _test():
        # To run from the command line: python build.py _test
        # This requires Python >=2.1 to run.
        import doctest, citdl
        return doctest.testmod(citdl)
    def _test_TclImportHandler():
        importer = TclImportHandler()
        importer.setCorePath()
        print importer._getPath()
        path = None
        #path = [r"D:\trentm\as\Apps\Komodo-devel\src"]
        for file in importer.genScannableFiles(path, skipRareImports=True):
            print "scannable Tcl file: '%s'" % file
    #_test()
    _test_TclImportHandler()

