#!python
# Copyright (c) 2004 Sophos Inc.
# See the file LICENSE.txt for licensing information.

"""Code Intelligence: common definitions"""

import os
import sys
import md5
import re
import stat
import time

import logging



#---- exceptions

class CodeIntelError(Exception):
    """Base Code Intelligence system error."""
    pass

class LanguageEngineError(CodeIntelError):
    """An error in a Code Intelligence system Language Engine."""
    #XXX Should add some relevant data to the exception. Perhaps
    #    the request should be passed in and this c'tor can extract
    #    data it wants to keep.  This could be used to facilitate
    #    submitting bug reports on our Language Engines.
    pass

class CIXError(CodeIntelError):
    """Code Intelligence XML error."""
    pass

class CIDBError(CodeIntelError):
    """Code Intelligence Database error."""
    pass

class CITDLError(CodeIntelError):
    """CITDL syntax error."""
    pass


#---- globals

# Possible return values from Manager.getCIDBUpgradeInfo() and
# Database.getUpgradeInfo()
(CIDB_UPGRADE_NOT_NECESSARY,
 CIDB_UPGRADE_NOT_POSSIBLE,
 CIDB_UPGRADE_NECESSARY) = range(3)

# Priorities at which scanning requests can be scheduled.
PRIORITY_CONTROL = 0        # Special sentinal priority to control scheduler
PRIORITY_IMMEDIATE = 1      # UI is requesting info on this file now
PRIORITY_CURRENT = 2        # UI requires info on this file soon
PRIORITY_OPEN = 3           # UI will likely require info on this file soon
PRIORITY_BACKGROUND = 4     # info may be needed sometime

# CIDB base type constants
BT_CLASSREF, BT_INTERFACEREF = range(2)
# CIDB symbol type constants
(ST_FUNCTION, ST_CLASS, ST_INTERFACE, ST_VARIABLE, ST_ARGUMENT) = range(5)
_symbolType2Name = {
    ST_FUNCTION: "function",
    ST_CLASS: "class",
    ST_INTERFACE: "interface",
    ST_VARIABLE: "variable",
    ST_ARGUMENT: "argument"
}


#---- public methods

def symbolType2Name(st):
    return _symbolType2Name[st]

# match 0x00-0x1f except TAB(0x09), LF(0x0A), and CR(0x0D)
_encre = re.compile('([\x00-\x08\x0b\x0c\x0e-\x1f])')
if sys.version_info >= (2, 3):
    charrefreplace = 'xmlcharrefreplace'
else:
    # Python 2.2 doesn't have 'xmlcharrefreplace'. Fallback to a
    # literal '?' -- this is better than failing outright.
    charrefreplace = 'replace'

def xmlencode(s):
    """Encode the given string for inclusion in a UTF-8 XML document.
    
    Specifically, illegal or unpresentable characters are encoded as
    XML character entities.
    """
    # As defined in the XML spec some of the character from 0x00 to 0x19
    # are not allowed in well-formed XML. We replace those with entity
    # references here.
    #   http://www.w3.org/TR/2000/REC-xml-20001006#charsets
    # (XXX It would be nice if Python has a codec for this. Perhaps we
    # should write one.)
    return _encre.sub(
               # replace with XML decimal char entity, e.g. '&#7;'
               lambda m: '&#%d;'%ord(m.group(1)),
               s.encode('utf-8', charrefreplace))

def xmlattrstr(attrs):
    """Construct an XML-safe attribute string from the given attributes
    
        "attrs" is a dictionary of attributes
    
    The returned attribute string includes a leading space, if necessary,
    so it is safe to use the string right after a tag name.
    """
    #XXX Should this be using 
    from xml.sax.saxutils import quoteattr
    s = ''
    names = attrs.keys()
    names.sort() # dump attrs sorted by key, not necessary but is more stable
    for name in names:
        s += ' %s=%s' % (name, quoteattr(str(attrs[name])))
    return s

def getPlatform():
    """Return the current platform name."""
    import re
    if sys.platform == "win32":
        plat = "win32-ix86"  #XXX what about "win64-ia64"?
    elif sys.platform.startswith("linux"):
        uname = os.uname()
        if uname[4] == "ia64":
            plat = "linux-ia64"
        elif re.match("i\d86", uname[4]):
            plat = "linux-ix86"
        elif uname[4] == "x86_64":
            plat = "linux-x86_64"
        else:
            raise CodeIntelError("unknown Linux architecture: '%s'" % uname[4])
    elif sys.platform.startswith("sunos"):
        uname = os.uname()
        if uname[4].startswith("sun4"):
            match = re.match(r"\d+\.(\d+)", uname[2])
            if uname[2] == "5.6":
                plat = "solaris-sparc"
            elif match:
                plat = "solaris%s-sparc" % match.group(1)
            else:
                raise CodeIntelError("unknown Solaris version: '%s'" % uname[2])
        else:
            raise CodeIntelError("unknown Solaris architecture: '%s'" % uname[4])
    elif sys.platform.startswith("hp-ux"):
        #XXX How to detect "PA2.0W" vs "PA1.1"? What are the backward
        #    compat issues here? What about HP-UX 11.x vs HP-UX 10.x
        #    compat issues?
        uname = os.uname()
        if uname[4] == "ia64":
            plat = "hpux-ia64"
        elif uname[4] == "9000/800":
            plat = "hpux-parisc"
        else:
            raise CodeIntelError("unknown HP-UX architecture: '%s'" % uname[4])
    elif sys.platform.startswith("aix"):
        # Determine processor type from 'uname -p'.
        o = os.popen('uname -p 2> /dev/null')
        processor = o.read().strip()
        o.close()

        uname = os.uname()
        plat = "aix%s-%s" % (uname[3], processor)
    return plat

def isUnsavedPath(path):
    """Return true if the given path is a special <Unsaved>\sub\path file."""
    tag = "<Unsaved>"
    length = len(tag)
    if path.startswith(tag) and (len(path)==length or path[length] in "\\/"):
        return True
    else:
        return False

_uriMatch = re.compile("^\w+://")
def canonicalizePath(path, normcase=True):
    r"""Return what CodeIntel considers a canonical version of the given path.
    
        "path" is the path to canonicalize.
        "normcase" (optional, default True) is a boolean indicating if the
            case should be normalized.
    
    "Special" paths are ones of the form "<Tag>\sub\path". Supported special
    path tags:
        <Unsaved>       Used when the given path isn't a real file: e.g.
                        unsaved document buffers.

    Raises a ValueError if it cannot be converted to a canonical path.
    
    >>> canonicalizePath(r"C:\Python22\Lib\os.py")  # normcase on Windows
    'c:\\python22\\lib\\os.py'
    >>> canonicalizePath(r"<Unsaved>\Python-1.py")
    '<Unsaved>\\python-1.py'
    >>> canonicalizePath("<Unsaved>")
    '<Unsaved>'
    >>> canonicalizePath("<Unsaved>\\")
    '<Unsaved>'
    >>> canonicalizePath("ftp://ftp.ActiveState.com/pub")
    'ftp://ftp.ActiveState.com/pub'
    """
    if path is None:
        raise ValueError("cannot canonicalize path, path is None")
    if path.startswith('<'):  # might be a special path
        first, rest = None, None
        for i in range(1, len(path)):
            if path[i] in "\\/":
                first, rest = path[:i], path[i+1:]
                break
        else:
            first, rest = path, None
        if first.endswith('>'):
            tag = first
            subpath = rest
            if tag == "<Unsaved>":
                pass # leave tag unchanged
            else:
                raise ValueError("unknown special path tag: %s" % tag)
            cpath = tag
            if subpath:
                subpath = os.path.normpath(subpath)
                if normcase:
                    subpath = os.path.normcase(subpath)
                cpath = os.path.join(cpath, subpath)
            return cpath
    if _uriMatch.match(path): # ftp://, koremote://
        #XXX Should we normcase() a UR[LI]
        return path
    else:
        cpath = os.path.normpath(os.path.abspath(path))
        if normcase:
            cpath = os.path.normcase(cpath)
        return cpath

def parseAttributes(attrStr=None):
    """Parse the given attributes string (from CIX) into an attribute dict."""
    attrs = {}
    if attrStr is not None:
        for token in attrStr.split():
            if '=' in token:
                key, value = token.split('=', 1)
            else:
                key, value = token, 1
            attrs[key] = value
    return attrs



#---- public classes

class ScanRequest:
    """A request to scan a file for code intel.
    
    A ScanRequest has the following properties:
        "id" is a unique ID for this request. It is assigned by the
            scheduler when the request is made.
        "path" is the full canonicalized path to the file to scan. The path
            is canonicalized when set in the constructor.
        "language" is the content-type of the file. This determines what
            "Language Engine(s)" are used to scan it. (XXX Specify allowable
            range of languages.)
        "priority" must be one of the PRIORITY_* priorities.
        "force" is a boolean indicating if a scan should be run even if
            the database is already up-to-date for this content.
        "content" is the file content. This can be explicitly given in the
            constructor. This is useful if the current content is not
            saved to disk or if it is difficult to retrieve the content.
            If not specified the loadContent() method will be used to
            get it from the filename.  Subclasses can override
            loadContent() if necessary.
        "md5sum" is the MD5 hexdigest of the content.  If already
            calculated it may be specified in the constructor. Otherwise
            this can be calculated as needed via the calculateMD5()
            method.
        "mtime" is the modified time of the file/content. If this is not
            given, it is determined lazily (if content is NOT specified) or
            defaults to the current time (if content IS specified).
        "scan_imports" is a boolean (default true) indicating that
            imports should be scheduled for scanning when this file is
            loaded into the database. 
    """
    def __init__(self, path, language, priority, force=0, content=None,
                 md5sum=None, mtime=None, scan_imports=True):
        self.id = None
        self.path = path
        self.language = language
        self.priority = priority
        self.force = force
        self.content = content
        if mtime:
            self.mtime = mtime
        elif content is not None:
            # Presumably if the content is being specified rather than having
            # the request's loadContent() determine it, then it is from an
            # editor buffer and may have changed just recently.
            self.mtime = int(time.time())
        else:
            self.mtime = None # determine lazily
        self.md5sum = md5sum
        self.scan_imports = scan_imports
    def __repr__(self):
        return "<ScanRequest id:%r, path:'%s'>" % (self.id, self.path)
    def loadContent(self):
        """If self.content is not set, load it from self.path.
        
        This also sets self.mtime, if necessary.
        This can raise an EnvironmentError if the file is not accessible.
        """
        if self.content is None:
            self.mtime = os.stat(self.path)[stat.ST_MTIME]
            fin = open(self.path, "r")
            try:
                self.content = fin.read()
            finally:
                fin.close()
    def calculateMD5(self):
        """Calculate and set self.md5sum if is it not already set."""
        if self.md5sum is None:
            self.loadContent()
            self.md5sum = md5.new(self.content).hexdigest()
    def getCanonicalPath(self):
        return canonicalizePath(self.path)




#---- self-test code

if __name__ == '__main__':
    def _test():
        import doctest, common
        return doctest.testmod(common)
    _test()
