#!/usr/bin/env python
# Copyright (c) 2000-2004 ActiveState Corp.
# See the file LICENSE.txt for licensing information.
#
#********************************************************************
# WARNING: Do not run this script directly. Run the main 
#          "install.sh" which will launch this script
#          properly.
#********************************************************************

"""
    install.sh - Komodo install script

    Usage:
        install.sh [options...]

    General Options:
        -h, --help          print this help and exit
        -v, --verbose       verbose output

        -I, --install-dir <dir>     specify install directory

    When called without arguments this script will interactively install
    Komodo. If the install dir is specified then Komodo will be
    installed without interaction.
"""

import sys
import os
import shutil
import socket
import tempfile
import getopt
import logging
import pprint
import stat
import re


class Error(Exception): pass



#---- globals

log = logging.getLogger("install")
gDefaultInstallDir = "/opt/Komodo-3.1"
gInstallLog = "koinstall.log"



#---- internal support routines

def _rmtreeOnError(rmFunction, filePath, excInfo):
    if excInfo[0] == OSError:
        # presuming because file is read-only
        os.chmod(filePath, 0777)
        rmFunction(filePath)

def _rmtree(dirname):
    import shutil
    shutil.rmtree(dirname, 0, _rmtreeOnError)

if sys.platform.startswith("win"):
    def _getSystemDrive():
        try:
            return os.environ["SystemDrive"]
        except KeyError:
            raise Error("'SystemDrive' environment variable is not set")


def _getMozCmdPrefix(installDir, userDataDir):
    """Return a command prefix to properly setup the environment to run a
    Komodo/Mozilla program.

        "installDir" is the path where Komodo is being installed.
        "userDataDir" is a directory to which to redirect Komodo's user data
            dir.

    Note, I am reproducing the environment that komodo.cpp sets up. Those
    must be kept in sync.
    """
    path = [ os.path.join(installDir, "python", "bin"),
             os.path.join(installDir, "Mozilla") ]
    if os.environ.has_key("PATH"):
        path += os.environ["PATH"].split(os.pathsep)
    ld_library_path = [ os.path.join(installDir, "Mozilla") ]
    #XXX Do we need this on LD_LIBRARY_PATH anymore?
    #       os.path.join(installDir, "Mozilla", "components")
    if os.environ.has_key("LD_LIBRARY_PATH"):
        ld_library_path += os.environ["LD_LIBRARY_PATH"].split(os.pathsep)
    pythonhome = os.path.join(installDir, "python")
    pythonpath = [ os.path.join(installDir, "Mozilla", "python") ]
    #XXX Do we need this on PYTHONPATH anymore?
    #       os.path.join(installDir, "python")
    prefix = "PATH=%s LD_LIBRARY_PATH=%s PYTHONHOME=%s PYTHONPATH=%s "\
             "KOMODO_HOSTNAME=%s KOMODO_USERDATADIR=%s %s"\
             % (os.pathsep.join(path), os.pathsep.join(ld_library_path),
                pythonhome, os.pathsep.join(pythonpath),
                socket.gethostname(), userDataDir,
                os.path.join(installDir, "Mozilla", "run-mozilla.sh"))
    return prefix


def _getDefaultInstallDir():
    default = gDefaultInstallDir
    if sys.platform.startswith("win") and\
       default.lower().find("%systemdrive%") != -1:
        default = re.compile("%SystemDrive%", re.I).sub(_getSystemDrive(),
                                                        default)
    return default

def _askYesNo(question, default="yes"):
    """Ask the user the given question and their answer.
    
    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is one of "yes" or "no".
    """
    valid = {"yes":"yes",   "y":"yes",  "ye":"yes",
             "no":"no",     "n":"no"}
    if default == None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise Error("invalid default answer: '%s'" % default)

    while 1:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        #sys.stdout.write('\n')
        if default is not None and choice == '':
            return default
        elif choice in valid.keys():
            return valid[choice]
        else:
            sys.stdout.write("Please repond with 'yes' or 'no' "\
                             "(or 'y' or 'n').\n")


def _validateInstallDir(installDir):
    if os.path.exists(installDir) and not os.path.isdir(installDir):
        raise Error("cannot install to '%s': exists and is not a directory"
                    % installDir)


def _isBinary(filename):
    """Return true iff the given filename is binary.

    Raises an EnvironmentError if the file does not exist or cannot be
    accessed.
    """
    fin = open(filename, 'rb')
    try:
        CHUNKSIZE = 1024
        while 1:
            chunk = fin.read(CHUNKSIZE)
            if '\0' in chunk: # found null byte
                return 1
            if len(chunk) < CHUNKSIZE:
                break # done
    finally:
        fin.close()
    return 0


def _relocateFilesInDir((buildPrefix, installPrefix), dname, fnames):
    if len(installPrefix) > len(buildPrefix):
        raise Error("cannot properly relocate dir-dependent binaries "
                    "because install path is too long (%d chars): you "
                    "must install ActivePython to a path with fewer than "
                    "%d characters" % (len(installPrefix), len(buildPrefix)))
    for fname in fnames:
        fpath = os.path.join(dname, fname)
        if os.path.isdir(fpath):
            continue

        # Determine if this file needs to be relocated.
        fin = open(fpath, 'rb')
        try:
            content = fin.read()
        finally:
            fin.close()
        isBinary = _isBinary(fpath)
        if isBinary:
            this = os.path.join(buildPrefix, "lib")
            that = os.path.join(installPrefix, "lib")
        else:
            this = buildPrefix
            that = installPrefix
        if content.find(this) == -1:
            continue

        # Relocate this file.
        log.debug("relocate '%s'", fpath)
        perm = stat.S_IMODE(os.stat(fpath).st_mode)
        if isBinary:
            if sys.platform.startswith("aix"):
                # On AIX the lib path _list_ is stored as one string, rather
                # than just the one path. This means that the integrity of
                # the path list must be maintained by separating with ':'.
                # We also change the remainder to all x's to ensure it is
                # a bogus path.
                that = os.path.join(installPrefix, "lib")\
                       + ':' + "x"*(len(buildPrefix)-len(installPrefix)-1)
                log.debug("replace (length %d)\n\t%s\nwith (length %d)\n\t%s",
                          len(this), this, len(that), that)
                content = content.replace(this, that)
            else:
                # Replace 'this' with 'that' in a null-terminated string.
                # Make sure to properly correct for trailing content in the
                # same string because:
                # - on HP-UX sometimes a full path to the shared lib is stored:
                #      <this>/libtcl8.4.sl\0
                # - on AIX a path _list_ is stored:
                #      <this>:other/lib/paths\0
                #   NOTE: This *should* work on AIX, AFAICT, but it does
                #   *not*.
                pattern = re.compile("%s([^\0]*)\0" % re.escape(this))
                def c_string_replace(match, before=this, after=that):
                    lendiff = len(before) - len(after)
                    s = after + match.group(1) + ("\0"*lendiff) + "\0"
                    # Encode nulls as '0' instead of '\x00' so one can see
                    # the before and after strings line up.
                    log.debug("replace (length %d)\n\t%s\nwith (length %d)\n\t%s",
                              len(match.group(0)),
                              repr(match.group(0)).replace("\\x00", '0'),
                              len(s),
                              repr(s).replace("\\x00", '0'))
                    return s
                content = pattern.sub(c_string_replace, content)
        else:
            log.debug("replace (length %d)\n\t%s\nwith (length %d)\n\t%s",
                      len(this), this, len(that), that)
            content = content.replace(this, that)
        # Sometimes get the following error. Avoid it by removing file first.
        #   IOError: [Errno 26] Text file busy: '$fpath'
        os.remove(fpath)
        fout = open(fpath, 'wb')
        try:
            fout.write(content)
        finally:
            fout.close()
        os.chmod(fpath, perm) # restore permissions


def _install(installDir, userDataDir):
    normInstallDir = os.path.normpath(os.path.expanduser(installDir))
    absInstallDir = os.path.abspath(normInstallDir)
    log.info("Installing ActiveState Komodo to '%s'...", normInstallDir)
    _validateInstallDir(absInstallDir)
    log.debug("redirect userDataDir to '%s'", userDataDir)
    log.debug("user environment: %s", pprint.pformat(dict(os.environ)))

    # copy the entire "Komodo" tree to the installDir
    if not os.path.exists(absInstallDir):
        os.makedirs(absInstallDir)
    import sh2
    sh2.cp(os.path.join(os.path.dirname(__file__), "INSTALLDIR", "*"),
           dstdir=absInstallDir, preserve=True,
           recursive=True)

    log.debug("Preparing internal Python...")
    import activestate
    pyInstallPrefix = os.path.join(absInstallDir, "python")
    pyBuildPrefix = activestate.prefix
    log.debug("relocating siloed Python from '%s' to '%s'", pyBuildPrefix,
              pyInstallPrefix)
    os.path.walk(pyInstallPrefix, _relocateFilesInDir,
                 (pyBuildPrefix, pyInstallPrefix))

    log.debug("pre-compile .py files in siloed Python")
    pyLibDir = os.path.join(absInstallDir, "python", "lib",
                            "python%s.%s" % sys.version_info[:2])
    # Skip .pyc one, just .pyo's are necessary because we always run
    # Python in optimized mode.
    cmds = ['%s -W "ignore:hex/oct constants" -O %s/compileall.py -q -x test/ %s'\
                % (sys.executable, pyLibDir, pyLibDir)]
    for cmd in cmds:
        log.debug("running '%s'" % cmd)
        retval = os.system(cmd)
        if retval:
            log.warn("error running '%s'" % cmd)

    # Configure siloed GTK2 libs (if any).
    log.debug("Configuring GTK2...")
    files = [os.path.join(absInstallDir, "Mozilla", "pango", "pangorc"),
             os.path.join(absInstallDir, "Mozilla", "pango", "pango.modules"),
             os.path.join(absInstallDir, "Mozilla", "gdk-pixbuf", "gdk-pixbuf.loaders")]
    token = "__PATH_TO_INSTALLED_KOMODO__"
    for file in files:
        if not os.path.exists(file):
            continue
        log.debug("fixing up '%s'", file)
        fin = open(file, 'r')
        try:
            content = fin.read()
        finally:
            fin.close()
        content = content.replace(token, absInstallDir)
        os.chmod(file, 0644)
        fout = open(file, 'w')
        try:
            fout.write(content)
        finally:
            fout.close()

    # run regxpcom
    # needed because:
    #    (1) If a priviledged user installs Komodo, regxpcom must be run
    #        with those privs to generate the chrome and XPCOM
    #        registries.
    #    (2) Komodo's license mechanism doesn't properly read the
    #        installed license if regxpcom is run in the same Mozilla
    #        session.
    log.info("Creating XPCOM registry...")
    # first clean out registry
    filesToClean = [
        os.path.join(normInstallDir, "Mozilla", "components", "compreg.dat"),
        os.path.join(normInstallDir, "Mozilla", "components", "xpti.dat")
    ]
    for file in filesToClean:
        if os.path.isfile(file):
            os.chmod(file, 0777)
            os.remove(file)
    # Would like to redirect stderr to /dev/null as well but this
    # causes, for some reason, case (2) from above to occur. Go figure.
    cmd = "(%s %s >/dev/null)" %\
          (_getMozCmdPrefix(normInstallDir, userDataDir),
           os.path.join(normInstallDir, "Mozilla", "regxpcom"))
    log.debug("running '%s'", cmd)
    retval = os.system(cmd)
    log.debug("RETVAL: %r", retval)

    print """
Komodo has been successfully installed to:
    %s
    
You might want to add 'komodo' to your PATH by adding the 
install dir to you PATH. Bash users can add the following
to their ~/.bashrc file:
    PATH=%s:$PATH
Or you could create a symbolic link to 'komodo':
    ln -s %s/komodo /usr/local/bin/komodo

Documentation is available in Komodo or on the web here:
    http://aspn.activestate.com/ASPN/docs/Komodo

Please send us any feedback you might have or log bugs here:
    komodo-feedback@ActiveState.com
    http://bugs.activestate.com/Komodo/

Thank you for using Komodo.
""" % (absInstallDir, absInstallDir, absInstallDir)



#---- main public functions

def interactiveInstall():
    default = _getDefaultInstallDir()
    sys.stdout.write("""\
Enter directory in which to install Komodo. Leave blank and
press 'Enter' to use the default [%s].
Install directory: """ % default)
    installDir = raw_input().strip()
    if not installDir:
        installDir = default

    norm = os.path.normpath(os.path.expanduser(installDir))
    if os.path.isdir(norm):
        sys.stdout.write("""
'%s' already exists. Installing over an existing
Komodo installation may have unexpected results. Are you
sure you would like to proceed with the installation?
""" % installDir)
        choice = _askYesNo("Proceed?", default="no")
        if choice == "yes":
            pass
        elif choice == "no":
            print "Aborting install."
            return
    elif os.path.exists(norm):
        raise Error("'%s' exists and is not a directory" % installDir)

    print
    install(installDir)

def install(installDir):
    # Redirect the "user data dir" to a temp location to avoid
    # the problem described in bug 32270 ("sudo ./install.sh" results in
    # interfering root-owned stuff in ~/.komodo).
    tempDir = os.path.join(tempfile.gettempdir(), "koinstall.%s" % os.getpid())
    try:
        return _install(installDir, tempDir)
    finally:
        if os.path.exists(tempDir):
            log.debug("removing temp user data dir: '%s'", tempDir)
            _rmtree(tempDir)



#---- mainline

def main(argv):
    # Setup logging.
    hdlr = logging.FileHandler(gInstallLog, mode='w')
    formatter = logging.Formatter('%(asctime)s: %(levelname)s: %(message)s')
    hdlr.setFormatter(formatter)
    log.addHandler(hdlr) 
    console = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s: %(message)s')
    console.setFormatter(formatter)
    console.setLevel(logging.INFO)
    log.addHandler(console) 
    log.setLevel(logging.DEBUG)

    # Parse options.
    try:
        opts, args = getopt.getopt(argv[1:], "VvhI:",
            ["version", "verbose", "help", "install-dir="])
    except getopt.GetoptError, ex:
        log.error(str(ex))
        log.error("Try `install.sh --help'.")
        return 1
    installDir = None
    for opt, optarg in opts:
        if opt in ("-h", "--help"):
            sys.stdout.write(__doc__)
            return
        elif opt in ("-V", "--version"):
            ver = '.'.join([str(part) for part in _version_])
            print "install %s" % ver
            return
        elif opt in ("-v", "--verbose"):
            console.setLevel(logging.DEBUG)
        elif opt in ("-I", "--install-dir"):
            installDir = optarg

    try:
        if installDir is None:
            interactiveInstall()
        else:
            install(installDir)
    except (EnvironmentError, Error), ex:
        log.error(str(ex))
        log.debug("exception info:", exc_info=True)
        return 1
    except KeyboardInterrupt:
        log.debug("user abort")
        pass

if __name__ == "__main__":
    __file__ == sys.argv[0]
    sys.exit( main(sys.argv) )
